import hashlib
import os
import platform
import subprocess
import sys
import warnings
from io import StringIO
from itertools import product
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import (
    Any,
    Callable,
    Dict,
    Hashable,
    List,
    Literal,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
    overload,
)

import numpy as np
from loguru import logger
from numpy.typing import NDArray
from typing_extensions import NotRequired, TypedDict, Unpack, assert_never, get_args

from bioimageio.core import __version__
from bioimageio.spec import (
    AnyDatasetDescr,
    AnyModelDescr,
    BioimageioCondaEnv,
    DatasetDescr,
    InvalidDescr,
    LatestResourceDescr,
    ModelDescr,
    ResourceDescr,
    ValidationContext,
    build_description,
    dump_description,
    get_conda_env,
    load_description,
    save_bioimageio_package,
)
from bioimageio.spec._description_impl import DISCOVER
from bioimageio.spec._internal.common_nodes import ResourceDescrBase
from bioimageio.spec._internal.io import is_yaml_value
from bioimageio.spec._internal.io_utils import read_yaml, write_yaml
from bioimageio.spec._internal.types import (
    AbsoluteTolerance,
    FormatVersionPlaceholder,
    MismatchedElementsPerMillion,
    RelativeTolerance,
)
from bioimageio.spec._internal.validation_context import get_validation_context
from bioimageio.spec.common import BioimageioYamlContent, PermissiveFileSource, Sha256
from bioimageio.spec.model import v0_4, v0_5
from bioimageio.spec.model.v0_5 import WeightsFormat
from bioimageio.spec.summary import (
    ErrorEntry,
    InstalledPackage,
    ValidationDetail,
    ValidationSummary,
    WarningEntry,
)

from ._prediction_pipeline import create_prediction_pipeline
from .axis import AxisId, BatchSize
from .common import MemberId, SupportedWeightsFormat
from .digest_spec import get_test_input_sample, get_test_output_sample
from .sample import Sample


class DeprecatedKwargs(TypedDict):
    absolute_tolerance: NotRequired[AbsoluteTolerance]
    relative_tolerance: NotRequired[RelativeTolerance]
    decimal: NotRequired[Optional[int]]


def enable_determinism(
    mode: Literal["seed_only", "full"] = "full",
    weight_formats: Optional[Sequence[SupportedWeightsFormat]] = None,
):
    """Seed and configure ML frameworks for maximum reproducibility.
    May degrade performance. Only recommended for testing reproducibility!

    Seed any random generators and (if **mode**=="full") request ML frameworks to use
    deterministic algorithms.

    Args:
        mode: determinism mode
            - 'seed_only' -- only set seeds, or
            - 'full' determinsm features (might degrade performance or throw exceptions)
        weight_formats: Limit deep learning importing deep learning frameworks
            based on weight_formats.
            E.g. this allows to avoid importing tensorflow when testing with pytorch.

    Notes:
        - **mode** == "full"  might degrade performance or throw exceptions.
        - Subsequent inference calls might still differ. Call before each function
          (sequence) that is expected to be reproducible.
        - Degraded performance: Use for testing reproducibility only!
        - Recipes:
            - [PyTorch](https://pytorch.org/docs/stable/notes/randomness.html)
            - [Keras](https://keras.io/examples/keras_recipes/reproducibility_recipes/)
            - [NumPy](https://numpy.org/doc/2.0/reference/random/generated/numpy.random.seed.html)
    """
    try:
        try:
            import numpy.random
        except ImportError:
            pass
        else:
            numpy.random.seed(0)
    except Exception as e:
        logger.debug(str(e))

    if (
        weight_formats is None
        or "pytorch_state_dict" in weight_formats
        or "torchscript" in weight_formats
    ):
        try:
            try:
                import torch
            except ImportError:
                pass
            else:
                _ = torch.manual_seed(0)
                torch.use_deterministic_algorithms(mode == "full")
        except Exception as e:
            logger.debug(str(e))

    if (
        weight_formats is None
        or "tensorflow_saved_model_bundle" in weight_formats
        or "keras_hdf5" in weight_formats
    ):
        try:
            os.environ["TF_ENABLE_ONEDNN_OPTS"] = "0"
            try:
                import tensorflow as tf  # pyright: ignore[reportMissingTypeStubs]
            except ImportError:
                pass
            else:
                tf.random.set_seed(0)
                if mode == "full":
                    tf.config.experimental.enable_op_determinism()
                # TODO: find possibility to switch it off again??
        except Exception as e:
            logger.debug(str(e))

    if weight_formats is None or "keras_hdf5" in weight_formats:
        try:
            try:
                import keras  # pyright: ignore[reportMissingTypeStubs]
            except ImportError:
                pass
            else:
                keras.utils.set_random_seed(0)
        except Exception as e:
            logger.debug(str(e))


def test_model(
    source: Union[v0_4.ModelDescr, v0_5.ModelDescr, PermissiveFileSource],
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[List[str]] = None,
    *,
    determinism: Literal["seed_only", "full"] = "seed_only",
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    **deprecated: Unpack[DeprecatedKwargs],
) -> ValidationSummary:
    """Test model inference"""
    return test_description(
        source,
        weight_format=weight_format,
        devices=devices,
        determinism=determinism,
        expected_type="model",
        sha256=sha256,
        stop_early=stop_early,
        **deprecated,
    )


def default_run_command(args: Sequence[str]):
    logger.info("running '{}'...", " ".join(args))
    _ = subprocess.run(args, shell=True, text=True, check=True)


def test_description(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Union[FormatVersionPlaceholder, str] = "discover",
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    determinism: Literal["seed_only", "full"] = "seed_only",
    expected_type: Optional[str] = None,
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    runtime_env: Union[
        Literal["currently-active", "as-described"], Path, BioimageioCondaEnv
    ] = ("currently-active"),
    run_command: Callable[[Sequence[str]], None] = default_run_command,
    **deprecated: Unpack[DeprecatedKwargs],
) -> ValidationSummary:
    """Test a bioimage.io resource dynamically,
    for example run prediction of test tensors for models.

    Args:
        source: model description source.
        weight_format: Weight format to test.
            Default: All weight formats present in **source**.
        devices: Devices to test with, e.g. 'cpu', 'cuda'.
            Default (may be weight format dependent): ['cuda'] if available, ['cpu'] otherwise.
        determinism: Modes to improve reproducibility of test outputs.
        expected_type: Assert an expected resource description `type`.
        sha256: Expected SHA256 value of **source**.
                (Ignored if **source** already is a loaded `ResourceDescr` object.)
        stop_early: Do not run further subtests after a failed one.
        runtime_env: (Experimental feature!) The Python environment to run the tests in
            - `"currently-active"`: Use active Python interpreter.
            - `"as-described"`: Use `bioimageio.spec.get_conda_env` to generate a conda
                environment YAML file based on the model weights description.
            - A `BioimageioCondaEnv` or a path to a conda environment YAML file.
                Note: The `bioimageio.core` dependency will be added automatically if not present.
        run_command: (Experimental feature!) Function to execute (conda) terminal commands in a subprocess.
            The function should raise an exception if the command fails.
            **run_command** is ignored if **runtime_env** is `"currently-active"`.
    """
    if runtime_env == "currently-active":
        rd = load_description_and_test(
            source,
            format_version=format_version,
            weight_format=weight_format,
            devices=devices,
            determinism=determinism,
            expected_type=expected_type,
            sha256=sha256,
            stop_early=stop_early,
            **deprecated,
        )
        return rd.validation_summary

    if runtime_env == "as-described":
        conda_env = None
    elif isinstance(runtime_env, (str, Path)):
        conda_env = BioimageioCondaEnv.model_validate(read_yaml(Path(runtime_env)))
    elif isinstance(runtime_env, BioimageioCondaEnv):
        conda_env = runtime_env
    else:
        assert_never(runtime_env)

    td_kwargs: Dict[str, Any] = (
        dict(ignore_cleanup_errors=True) if sys.version_info >= (3, 10) else {}
    )
    with TemporaryDirectory(**td_kwargs) as _d:
        working_dir = Path(_d)
        if isinstance(source, (dict, ResourceDescrBase)):
            file_source = save_bioimageio_package(
                source, output_path=working_dir / "package.zip"
            )
        else:
            file_source = source

        return _test_in_env(
            file_source,
            working_dir=working_dir,
            weight_format=weight_format,
            conda_env=conda_env,
            devices=devices,
            determinism=determinism,
            expected_type=expected_type,
            sha256=sha256,
            stop_early=stop_early,
            run_command=run_command,
            **deprecated,
        )


def _test_in_env(
    source: PermissiveFileSource,
    *,
    working_dir: Path,
    weight_format: Optional[SupportedWeightsFormat],
    conda_env: Optional[BioimageioCondaEnv],
    devices: Optional[Sequence[str]],
    determinism: Literal["seed_only", "full"],
    run_command: Callable[[Sequence[str]], None],
    stop_early: bool,
    expected_type: Optional[str],
    sha256: Optional[Sha256],
    **deprecated: Unpack[DeprecatedKwargs],
) -> ValidationSummary:
    descr = load_description(source)

    if not isinstance(descr, (v0_4.ModelDescr, v0_5.ModelDescr)):
        raise NotImplementedError("Not yet implemented for non-model resources")

    if weight_format is None:
        all_present_wfs = [
            wf for wf in get_args(WeightsFormat) if getattr(descr.weights, wf)
        ]
        ignore_wfs = [wf for wf in all_present_wfs if wf in ["tensorflow_js"]]
        logger.info(
            "Found weight formats {}. Start testing all{}...",
            all_present_wfs,
            f" (except: {', '.join(ignore_wfs)}) " if ignore_wfs else "",
        )
        summary = _test_in_env(
            source,
            working_dir=working_dir / all_present_wfs[0],
            weight_format=all_present_wfs[0],
            devices=devices,
            determinism=determinism,
            conda_env=conda_env,
            run_command=run_command,
            expected_type=expected_type,
            sha256=sha256,
            stop_early=stop_early,
            **deprecated,
        )
        for wf in all_present_wfs[1:]:
            additional_summary = _test_in_env(
                source,
                working_dir=working_dir / wf,
                weight_format=wf,
                devices=devices,
                determinism=determinism,
                conda_env=conda_env,
                run_command=run_command,
                expected_type=expected_type,
                sha256=sha256,
                stop_early=stop_early,
                **deprecated,
            )
            for d in additional_summary.details:
                # TODO: filter reduntant details; group details
                summary.add_detail(d)
        return summary

    if weight_format == "pytorch_state_dict":
        wf = descr.weights.pytorch_state_dict
    elif weight_format == "torchscript":
        wf = descr.weights.torchscript
    elif weight_format == "keras_hdf5":
        wf = descr.weights.keras_hdf5
    elif weight_format == "onnx":
        wf = descr.weights.onnx
    elif weight_format == "tensorflow_saved_model_bundle":
        wf = descr.weights.tensorflow_saved_model_bundle
    elif weight_format == "tensorflow_js":
        raise RuntimeError(
            "testing 'tensorflow_js' is not supported by bioimageio.core"
        )
    else:
        assert_never(weight_format)

    assert wf is not None
    if conda_env is None:
        conda_env = get_conda_env(entry=wf)

    # remove name as we crate a name based on the env description hash value
    conda_env.name = None

    dumped_env = conda_env.model_dump(mode="json", exclude_none=True)
    if not is_yaml_value(dumped_env):
        raise ValueError(f"Failed to dump conda env to valid YAML {conda_env}")

    env_io = StringIO()
    write_yaml(dumped_env, file=env_io)
    encoded_env = env_io.getvalue().encode()
    env_name = hashlib.sha256(encoded_env).hexdigest()

    try:
        run_command(["where" if platform.system() == "Windows" else "which", "conda"])
    except Exception as e:
        raise RuntimeError("Conda not available") from e

    try:
        run_command(["conda", "activate", env_name])
    except Exception:
        path = working_dir / "env.yaml"
        try:
            _ = path.write_bytes(encoded_env)
            logger.debug("written conda env to {}", path)
            run_command(
                ["conda", "env", "create", f"--file={path}", f"--name={env_name}"]
            )
            run_command(["conda", "activate", env_name])
        except Exception as e:
            summary = descr.validation_summary
            summary.add_detail(
                ValidationDetail(
                    name="Conda environment creation",
                    status="failed",
                    loc=("weights", weight_format),
                    recommended_env=conda_env,
                    errors=[
                        ErrorEntry(
                            loc=("weights", weight_format),
                            msg=str(e),
                            type="conda",
                            with_traceback=True,
                        )
                    ],
                )
            )
            return summary

    working_dir.mkdir(parents=True, exist_ok=True)
    summary_path = working_dir / "summary.json"
    assert not summary_path.exists(), "Summary file already exists"
    cmd = []
    cmd_error = None
    for summary_path_arg_name in ("summary", "summary-path"):
        try:
            run_command(
                cmd := (
                    [
                        "conda",
                        "run",
                        "-n",
                        env_name,
                        "bioimageio",
                        "test",
                        str(source),
                        f"--{summary_path_arg_name}={summary_path.as_posix()}",
                        f"--determinism={determinism}",
                    ]
                    + ([f"--expected-type={expected_type}"] if expected_type else [])
                    + (["--stop-early"] if stop_early else [])
                )
            )
        except Exception as e:
            cmd_error = f"Failed to run command '{' '.join(cmd)}': {e}."

        if summary_path.exists():
            break
    else:
        if cmd_error is not None:
            logger.warning(cmd_error)

        return ValidationSummary(
            name="calling bioimageio test command",
            source_name=str(source),
            status="failed",
            type="unknown",
            format_version="unknown",
            details=[
                ValidationDetail(
                    name="run 'bioimageio test'",
                    errors=[
                        ErrorEntry(
                            loc=(),
                            type="bioimageio cli",
                            msg=f"test command '{' '.join(cmd)}' did not produce a summary file at {summary_path}",
                        )
                    ],
                    status="failed",
                )
            ],
            env=set(),
        )

    return ValidationSummary.load_json(summary_path)


@overload
def load_description_and_test(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Literal["latest"],
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    determinism: Literal["seed_only", "full"] = "seed_only",
    expected_type: Literal["model"],
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    **deprecated: Unpack[DeprecatedKwargs],
) -> Union[ModelDescr, InvalidDescr]: ...


@overload
def load_description_and_test(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Literal["latest"],
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    determinism: Literal["seed_only", "full"] = "seed_only",
    expected_type: Literal["dataset"],
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    **deprecated: Unpack[DeprecatedKwargs],
) -> Union[DatasetDescr, InvalidDescr]: ...


@overload
def load_description_and_test(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Literal["latest"],
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    determinism: Literal["seed_only", "full"] = "seed_only",
    expected_type: Optional[str] = None,
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    **deprecated: Unpack[DeprecatedKwargs],
) -> Union[LatestResourceDescr, InvalidDescr]: ...


@overload
def load_description_and_test(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    determinism: Literal["seed_only", "full"] = "seed_only",
    expected_type: Literal["model"],
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    **deprecated: Unpack[DeprecatedKwargs],
) -> Union[AnyModelDescr, InvalidDescr]: ...


@overload
def load_description_and_test(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    determinism: Literal["seed_only", "full"] = "seed_only",
    expected_type: Literal["dataset"],
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    **deprecated: Unpack[DeprecatedKwargs],
) -> Union[AnyDatasetDescr, InvalidDescr]: ...


@overload
def load_description_and_test(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    determinism: Literal["seed_only", "full"] = "seed_only",
    expected_type: Optional[str] = None,
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    **deprecated: Unpack[DeprecatedKwargs],
) -> Union[ResourceDescr, InvalidDescr]: ...


def load_description_and_test(
    source: Union[ResourceDescr, PermissiveFileSource, BioimageioYamlContent],
    *,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    weight_format: Optional[SupportedWeightsFormat] = None,
    devices: Optional[Sequence[str]] = None,
    determinism: Literal["seed_only", "full"] = "seed_only",
    expected_type: Optional[str] = None,
    sha256: Optional[Sha256] = None,
    stop_early: bool = True,
    **deprecated: Unpack[DeprecatedKwargs],
) -> Union[ResourceDescr, InvalidDescr]:
    """Test a bioimage.io resource dynamically,
    for example run prediction of test tensors for models.

    See `test_description` for more details.

    Returns:
        A (possibly invalid) resource description object
        with a populated `.validation_summary` attribute.
    """
    if isinstance(source, ResourceDescrBase):
        root = source.root
        file_name = source.file_name
        if (
            (
                format_version
                not in (
                    DISCOVER,
                    source.format_version,
                    ".".join(source.format_version.split(".")[:2]),
                )
            )
            or (c := source.validation_summary.details[0].context) is None
            or not c.perform_io_checks
        ):
            logger.debug(
                "deserializing source to ensure we validate and test using format {} and perform io checks",
                format_version,
            )
            source = dump_description(source)
    else:
        root = Path()
        file_name = None

    if isinstance(source, ResourceDescrBase):
        rd = source
    elif isinstance(source, dict):
        # check context for a given root; default to root of source
        context = get_validation_context(
            ValidationContext(root=root, file_name=file_name)
        ).replace(
            perform_io_checks=True  # make sure we perform io checks though
        )

        rd = build_description(
            source,
            format_version=format_version,
            context=context,
        )
    else:
        rd = load_description(
            source, format_version=format_version, sha256=sha256, perform_io_checks=True
        )

    rd.validation_summary.env.add(
        InstalledPackage(name="bioimageio.core", version=__version__)
    )

    if expected_type is not None:
        _test_expected_resource_type(rd, expected_type)

    if isinstance(rd, (v0_4.ModelDescr, v0_5.ModelDescr)):
        if weight_format is None:
            weight_formats: List[SupportedWeightsFormat] = [
                w for w, we in rd.weights if we is not None
            ]  # pyright: ignore[reportAssignmentType]
        else:
            weight_formats = [weight_format]

        enable_determinism(determinism, weight_formats=weight_formats)
        for w in weight_formats:
            _test_model_inference(rd, w, devices, stop_early=stop_early, **deprecated)
            if stop_early and rd.validation_summary.status == "failed":
                break

            if not isinstance(rd, v0_4.ModelDescr):
                _test_model_inference_parametrized(
                    rd, w, devices, stop_early=stop_early
                )
                if stop_early and rd.validation_summary.status == "failed":
                    break

    # TODO: add execution of jupyter notebooks
    # TODO: add more tests

    if rd.validation_summary.status == "valid-format":
        rd.validation_summary.status = "passed"

    return rd


def _get_tolerance(
    model: Union[v0_4.ModelDescr, v0_5.ModelDescr],
    wf: SupportedWeightsFormat,
    m: MemberId,
    **deprecated: Unpack[DeprecatedKwargs],
) -> Tuple[RelativeTolerance, AbsoluteTolerance, MismatchedElementsPerMillion]:
    if isinstance(model, v0_5.ModelDescr):
        applicable = v0_5.ReproducibilityTolerance()

        # check legacy test kwargs for weight format specific tolerance
        if model.config.bioimageio.model_extra is not None:
            for weights_format, test_kwargs in model.config.bioimageio.model_extra.get(
                "test_kwargs", {}
            ).items():
                if wf == weights_format:
                    applicable = v0_5.ReproducibilityTolerance(
                        relative_tolerance=test_kwargs.get("relative_tolerance", 1e-3),
                        absolute_tolerance=test_kwargs.get("absolute_tolerance", 1e-4),
                    )
                    break

        # check for weights format and output tensor specific tolerance
        for a in model.config.bioimageio.reproducibility_tolerance:
            if (not a.weights_formats or wf in a.weights_formats) and (
                not a.output_ids or m in a.output_ids
            ):
                applicable = a
                break

        rtol = applicable.relative_tolerance
        atol = applicable.absolute_tolerance
        mismatched_tol = applicable.mismatched_elements_per_million
    elif (decimal := deprecated.get("decimal")) is not None:
        warnings.warn(
            "The argument `decimal` has been deprecated in favour of"
            + " `relative_tolerance` and `absolute_tolerance`, with different"
            + " validation logic, using `numpy.testing.assert_allclose, see"
            + " 'https://numpy.org/doc/stable/reference/generated/"
            + " numpy.testing.assert_allclose.html'. Passing a value for `decimal`"
            + " will cause validation to revert to the old behaviour."
        )
        atol = 1.5 * 10 ** (-decimal)
        rtol = 0
        mismatched_tol = 0
    else:
        # use given (deprecated) test kwargs
        atol = deprecated.get("absolute_tolerance", 1e-5)
        rtol = deprecated.get("relative_tolerance", 1e-3)
        mismatched_tol = 0

    return rtol, atol, mismatched_tol


def _test_model_inference(
    model: Union[v0_4.ModelDescr, v0_5.ModelDescr],
    weight_format: SupportedWeightsFormat,
    devices: Optional[Sequence[str]],
    stop_early: bool,
    **deprecated: Unpack[DeprecatedKwargs],
) -> None:
    test_name = f"Reproduce test outputs from test inputs ({weight_format})"
    logger.debug("starting '{}'", test_name)
    error_entries: List[ErrorEntry] = []
    warning_entries: List[WarningEntry] = []

    def add_error_entry(msg: str, with_traceback: bool = False):
        error_entries.append(
            ErrorEntry(
                loc=("weights", weight_format),
                msg=msg,
                type="bioimageio.core",
                with_traceback=with_traceback,
            )
        )

    def add_warning_entry(msg: str):
        warning_entries.append(
            WarningEntry(
                loc=("weights", weight_format),
                msg=msg,
                type="bioimageio.core",
            )
        )

    try:
        test_input = get_test_input_sample(model)
        expected = get_test_output_sample(model)

        with create_prediction_pipeline(
            bioimageio_model=model, devices=devices, weight_format=weight_format
        ) as prediction_pipeline:
            results = prediction_pipeline.predict_sample_without_blocking(test_input)

        if len(results.members) != len(expected.members):
            add_error_entry(
                f"Expected {len(expected.members)} outputs, but got {len(results.members)}"
            )

        else:
            for m, expected in expected.members.items():
                actual = results.members.get(m)
                if actual is None:
                    add_error_entry("Output tensors for test case may not be None")
                    if stop_early:
                        break
                    else:
                        continue

                if actual.dims != (dims := expected.dims):
                    add_error_entry(
                        f"Output '{m}' has dims {actual.dims}, but expected {expected.dims}"
                    )
                    if stop_early:
                        break
                    else:
                        continue

                if actual.tagged_shape != expected.tagged_shape:
                    add_error_entry(
                        f"Output '{m}' has shape {actual.tagged_shape}, but expected {expected.tagged_shape}"
                    )
                    if stop_early:
                        break
                    else:
                        continue

                expected_np = expected.data.to_numpy().astype(np.float32)
                del expected
                actual_np: NDArray[Any] = actual.data.to_numpy().astype(np.float32)
                del actual

                rtol, atol, mismatched_tol = _get_tolerance(
                    model, wf=weight_format, m=m, **deprecated
                )
                rtol_value = rtol * abs(expected_np)
                abs_diff = abs(actual_np - expected_np)
                mismatched = abs_diff > atol + rtol_value
                mismatched_elements = mismatched.sum().item()
                if not mismatched_elements:
                    continue

                mismatched_ppm = mismatched_elements / expected_np.size * 1e6
                abs_diff[~mismatched] = 0  # ignore non-mismatched elements

                r_max_idx_flat = (
                    r_diff := (abs_diff / (abs(expected_np) + 1e-6))
                ).argmax()
                r_max_idx = np.unravel_index(r_max_idx_flat, r_diff.shape)
                r_max = r_diff[r_max_idx].item()
                r_actual = actual_np[r_max_idx].item()
                r_expected = expected_np[r_max_idx].item()

                # Calculate the max absolute difference with the relative tolerance subtracted
                abs_diff_wo_rtol: NDArray[np.float32] = (abs_diff - rtol_value).max(
                    initial=0
                )
                a_max_idx = np.unravel_index(
                    abs_diff_wo_rtol.argmax(), abs_diff_wo_rtol.shape
                )

                a_max = abs_diff[a_max_idx].item()
                a_actual = actual_np[a_max_idx].item()
                a_expected = expected_np[a_max_idx].item()

                msg = (
                    f"Output '{m}' disagrees with {mismatched_elements} of"
                    + f" {expected_np.size} expected values"
                    + f" ({mismatched_ppm:.1f} ppm)."
                    + f"\n Max relative difference: {r_max:.2e}"
                    + rf" (= \|{r_actual:.2e} - {r_expected:.2e}\|/\|{r_expected:.2e} + 1e-6\|)"
                    + f" at {dict(zip(dims, r_max_idx))}"
                    + f"\n Max absolute difference not accounted for by relative tolerance: {a_max:.2e}"
                    + rf" (= \|{a_actual:.7e} - {a_expected:.7e}\|) at {dict(zip(dims, a_max_idx))}"
                )
                if mismatched_ppm > mismatched_tol:
                    add_error_entry(msg)
                    if stop_early:
                        break
                else:
                    add_warning_entry(msg)

    except Exception as e:
        if get_validation_context().raise_errors:
            raise e

        add_error_entry(str(e), with_traceback=True)

    model.validation_summary.add_detail(
        ValidationDetail(
            name=test_name,
            loc=("weights", weight_format),
            status="failed" if error_entries else "passed",
            recommended_env=get_conda_env(entry=dict(model.weights)[weight_format]),
            errors=error_entries,
            warnings=warning_entries,
        )
    )


def _test_model_inference_parametrized(
    model: v0_5.ModelDescr,
    weight_format: SupportedWeightsFormat,
    devices: Optional[Sequence[str]],
    *,
    stop_early: bool,
) -> None:
    if not any(
        isinstance(a.size, v0_5.ParameterizedSize)
        for ipt in model.inputs
        for a in ipt.axes
    ):
        # no parameterized sizes => set n=0
        ns: Set[v0_5.ParameterizedSize_N] = {0}
    else:
        ns = {0, 1, 2}

    given_batch_sizes = {
        a.size
        for ipt in model.inputs
        for a in ipt.axes
        if isinstance(a, v0_5.BatchAxis)
    }
    if given_batch_sizes:
        batch_sizes = {gbs for gbs in given_batch_sizes if gbs is not None}
        if not batch_sizes:
            # only arbitrary batch sizes
            batch_sizes = {1, 2}
    else:
        # no batch axis
        batch_sizes = {1}

    test_cases: Set[Tuple[BatchSize, v0_5.ParameterizedSize_N]] = {
        (b, n) for b, n in product(sorted(batch_sizes), sorted(ns))
    }
    logger.info(
        "Testing inference with '{}' for {} different inputs (B, N): {}",
        weight_format,
        len(test_cases),
        test_cases,
    )

    def generate_test_cases():
        tested: Set[Hashable] = set()

        def get_ns(n: int):
            return {
                (t.id, a.id): n
                for t in model.inputs
                for a in t.axes
                if isinstance(a.size, v0_5.ParameterizedSize)
            }

        for batch_size, n in sorted(test_cases):
            input_target_sizes, expected_output_sizes = model.get_axis_sizes(
                get_ns(n), batch_size=batch_size
            )
            hashable_target_size = tuple(
                (k, input_target_sizes[k]) for k in sorted(input_target_sizes)
            )
            if hashable_target_size in tested:
                continue
            else:
                tested.add(hashable_target_size)

            resized_test_inputs = Sample(
                members={
                    t.id: (
                        test_input.members[t.id].resize_to(
                            {
                                aid: s
                                for (tid, aid), s in input_target_sizes.items()
                                if tid == t.id
                            },
                        )
                    )
                    for t in model.inputs
                },
                stat=test_input.stat,
                id=test_input.id,
            )
            expected_output_shapes = {
                t.id: {
                    aid: s
                    for (tid, aid), s in expected_output_sizes.items()
                    if tid == t.id
                }
                for t in model.outputs
            }
            yield n, batch_size, resized_test_inputs, expected_output_shapes

    try:
        test_input = get_test_input_sample(model)

        with create_prediction_pipeline(
            bioimageio_model=model, devices=devices, weight_format=weight_format
        ) as prediction_pipeline:
            for n, batch_size, inputs, exptected_output_shape in generate_test_cases():
                error: Optional[str] = None
                result = prediction_pipeline.predict_sample_without_blocking(inputs)
                if len(result.members) != len(exptected_output_shape):
                    error = (
                        f"Expected {len(exptected_output_shape)} outputs,"
                        + f" but got {len(result.members)}"
                    )

                else:
                    for m, exp in exptected_output_shape.items():
                        res = result.members.get(m)
                        if res is None:
                            error = "Output tensors may not be None for test case"
                            break

                        diff: Dict[AxisId, int] = {}
                        for a, s in res.sizes.items():
                            if isinstance((e_aid := exp[AxisId(a)]), int):
                                if s != e_aid:
                                    diff[AxisId(a)] = s
                            elif (
                                s < e_aid.min or e_aid.max is not None and s > e_aid.max
                            ):
                                diff[AxisId(a)] = s
                        if diff:
                            error = (
                                f"(n={n}) Expected output shape {exp},"
                                + f" but got {res.sizes} (diff: {diff})"
                            )
                            break

                model.validation_summary.add_detail(
                    ValidationDetail(
                        name=f"Run {weight_format} inference for inputs with"
                        + f" batch_size: {batch_size} and size parameter n: {n}",
                        loc=("weights", weight_format),
                        status="passed" if error is None else "failed",
                        errors=(
                            []
                            if error is None
                            else [
                                ErrorEntry(
                                    loc=("weights", weight_format),
                                    msg=error,
                                    type="bioimageio.core",
                                )
                            ]
                        ),
                    )
                )
                if stop_early and error is not None:
                    break
    except Exception as e:
        if get_validation_context().raise_errors:
            raise e

        model.validation_summary.add_detail(
            ValidationDetail(
                name=f"Run {weight_format} inference for parametrized inputs",
                status="failed",
                loc=("weights", weight_format),
                errors=[
                    ErrorEntry(
                        loc=("weights", weight_format),
                        msg=str(e),
                        type="bioimageio.core",
                        with_traceback=True,
                    )
                ],
            )
        )


def _test_expected_resource_type(
    rd: Union[InvalidDescr, ResourceDescr], expected_type: str
):
    has_expected_type = rd.type == expected_type
    rd.validation_summary.details.append(
        ValidationDetail(
            name="Has expected resource type",
            status="passed" if has_expected_type else "failed",
            loc=("type",),
            errors=(
                []
                if has_expected_type
                else [
                    ErrorEntry(
                        loc=("type",),
                        type="type",
                        msg=f"Expected type {expected_type}, found {rd.type}",
                    )
                ]
            ),
        )
    )


# TODO: Implement `debug_model()`
# def debug_model(
#     model_rdf: Union[RawResourceDescr, ResourceDescr, URI, Path, str],
#     *,
#     weight_format: Optional[WeightsFormat] = None,
#     devices: Optional[List[str]] = None,
# ):
#     """Run the model test and return dict with inputs, results, expected results and intermediates.

#     Returns dict with tensors "inputs", "inputs_processed", "outputs_raw", "outputs", "expected" and "diff".
#     """
#     inputs_raw: Optional = None
#     inputs_processed: Optional = None
#     outputs_raw: Optional = None
#     outputs: Optional = None
#     expected: Optional = None
#     diff: Optional = None

#     model = load_description(
#         model_rdf, weights_priority_order=None if weight_format is None else [weight_format]
#     )
#     if not isinstance(model, Model):
#         raise ValueError(f"Not a bioimageio.model: {model_rdf}")

#     prediction_pipeline = create_prediction_pipeline(
#         bioimageio_model=model, devices=devices, weight_format=weight_format
#     )
#     inputs = [
#         xr.DataArray(load_array(str(in_path)), dims=input_spec.axes)
#         for in_path, input_spec in zip(model.test_inputs, model.inputs)
#     ]
#     input_dict = {input_spec.name: input for input_spec, input in zip(model.inputs, inputs)}

#     # keep track of the non-processed inputs
#     inputs_raw = [deepcopy(input) for input in inputs]

#     computed_measures = {}

#     prediction_pipeline.apply_preprocessing(input_dict, computed_measures)
#     inputs_processed = list(input_dict.values())
#     outputs_raw = prediction_pipeline.predict(*inputs_processed)
#     output_dict = {output_spec.name: deepcopy(output) for output_spec, output in zip(model.outputs, outputs_raw)}
#     prediction_pipeline.apply_postprocessing(output_dict, computed_measures)
#     outputs = list(output_dict.values())

#     if isinstance(outputs, (np.ndarray, xr.DataArray)):
#         outputs = [outputs]

#     expected = [
#         xr.DataArray(load_array(str(out_path)), dims=output_spec.axes)
#         for out_path, output_spec in zip(model.test_outputs, model.outputs)
#     ]
#     if len(outputs) != len(expected):
#         error = f"Number of outputs and number of expected outputs disagree: {len(outputs)} != {len(expected)}"
#         print(error)
#     else:
#         diff = []
#         for res, exp in zip(outputs, expected):
#             diff.append(res - exp)

#     return {
#         "inputs": inputs_raw,
#         "inputs_processed": inputs_processed,
#         "outputs_raw": outputs_raw,
#         "outputs": outputs,
#         "expected": expected,
#         "diff": diff,
#     }



# pyright: reportUnknownVariableType=false
import gc
import warnings
from typing import Any, List, Optional, Sequence, Union

import torch
from numpy.typing import NDArray

from bioimageio.spec.model import v0_4, v0_5

from ..model_adapters import ModelAdapter
from ..utils._type_guards import is_list, is_tuple


class TorchscriptModelAdapter(ModelAdapter):
    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ):
        super().__init__(model_description=model_description)
        if model_description.weights.torchscript is None:
            raise ValueError(
                f"No torchscript weights found for model {model_description.name}"
            )

        if devices is None:
            self.devices = ["cuda" if torch.cuda.is_available() else "cpu"]
        else:
            self.devices = [torch.device(d) for d in devices]

        if len(self.devices) > 1:
            warnings.warn(
                "Multiple devices for single torchscript model not yet implemented"
            )

        weight_reader = model_description.weights.torchscript.get_reader()
        self._model = torch.jit.load(weight_reader)

        self._model.to(self.devices[0])
        self._model = self._model.eval()

    def _forward_impl(
        self, input_arrays: Sequence[Optional[NDArray[Any]]]
    ) -> List[Optional[NDArray[Any]]]:

        with torch.no_grad():
            torch_tensor = [
                None if a is None else torch.from_numpy(a).to(self.devices[0])
                for a in input_arrays
            ]
            output: Any = self._model.forward(*torch_tensor)
            if is_list(output) or is_tuple(output):
                output_seq: Sequence[Any] = output
            else:
                output_seq = [output]

            return [
                (
                    None
                    if r is None
                    else r.cpu().numpy() if isinstance(r, torch.Tensor) else r
                )
                for r in output_seq
            ]

    def unload(self) -> None:
        self._devices = None
        del self._model
        _ = gc.collect()  # deallocate memory
        torch.cuda.empty_cache()  # release reserved memory



import warnings
from abc import ABC, abstractmethod
from typing import (
    Any,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    final,
)

from exceptiongroup import ExceptionGroup
from numpy.typing import NDArray
from typing_extensions import assert_never

from bioimageio.spec.model import AnyModelDescr, v0_4, v0_5

from ..common import SupportedWeightsFormat
from ..digest_spec import get_axes_infos, get_member_ids
from ..sample import Sample, SampleBlock, SampleBlockWithOrigin
from ..tensor import Tensor

# Known weight formats in order of priority
# First match wins
DEFAULT_WEIGHT_FORMAT_PRIORITY_ORDER: Tuple[SupportedWeightsFormat, ...] = (
    "pytorch_state_dict",
    "tensorflow_saved_model_bundle",
    "torchscript",
    "onnx",
    "keras_hdf5",
)


class ModelAdapter(ABC):
    """
    Represents model *without* any preprocessing or postprocessing.

    ```
    from bioimageio.core import load_description

    model = load_description(...)

    # option 1:
    adapter = ModelAdapter.create(model)
    adapter.forward(...)
    adapter.unload()

    # option 2:
    with ModelAdapter.create(model) as adapter:
        adapter.forward(...)
    ```
    """

    def __init__(self, model_description: AnyModelDescr):
        super().__init__()
        self._model_descr = model_description
        self._input_ids = get_member_ids(model_description.inputs)
        self._output_ids = get_member_ids(model_description.outputs)
        self._input_axes = [
            tuple(a.id for a in get_axes_infos(t)) for t in model_description.inputs
        ]
        self._output_axes = [
            tuple(a.id for a in get_axes_infos(t)) for t in model_description.outputs
        ]
        if isinstance(model_description, v0_4.ModelDescr):
            self._input_is_optional = [False] * len(model_description.inputs)
        else:
            self._input_is_optional = [ipt.optional for ipt in model_description.inputs]

    @final
    @classmethod
    def create(
        cls,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        *,
        devices: Optional[Sequence[str]] = None,
        weight_format_priority_order: Optional[Sequence[SupportedWeightsFormat]] = None,
    ):
        """
        Creates model adapter based on the passed spec
        Note: All specific adapters should happen inside this function to prevent different framework
        initializations interfering with each other
        """
        if not isinstance(model_description, (v0_4.ModelDescr, v0_5.ModelDescr)):
            raise TypeError(
                f"expected v0_4.ModelDescr or v0_5.ModelDescr, but got {type(model_description)}"
            )

        weights = model_description.weights
        errors: List[Exception] = []
        weight_format_priority_order = (
            DEFAULT_WEIGHT_FORMAT_PRIORITY_ORDER
            if weight_format_priority_order is None
            else weight_format_priority_order
        )
        # limit weight formats to the ones present
        weight_format_priority_order_present: Sequence[SupportedWeightsFormat] = [
            w for w in weight_format_priority_order if getattr(weights, w) is not None
        ]
        if not weight_format_priority_order_present:
            raise ValueError(
                f"None of the specified weight formats ({weight_format_priority_order}) is present ({weight_format_priority_order_present})"
            )

        for wf in weight_format_priority_order_present:
            if wf == "pytorch_state_dict":
                assert weights.pytorch_state_dict is not None
                try:
                    from .pytorch_backend import PytorchModelAdapter

                    return PytorchModelAdapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append(e)
            elif wf == "tensorflow_saved_model_bundle":
                assert weights.tensorflow_saved_model_bundle is not None
                try:
                    from .tensorflow_backend import create_tf_model_adapter

                    return create_tf_model_adapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append(e)
            elif wf == "onnx":
                assert weights.onnx is not None
                try:
                    from .onnx_backend import ONNXModelAdapter

                    return ONNXModelAdapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append(e)
            elif wf == "torchscript":
                assert weights.torchscript is not None
                try:
                    from .torchscript_backend import TorchscriptModelAdapter

                    return TorchscriptModelAdapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append(e)
            elif wf == "keras_hdf5":
                assert weights.keras_hdf5 is not None
                # keras can either be installed as a separate package or used as part of tensorflow
                # we try to first import the keras model adapter using the separate package and,
                # if it is not available, try to load the one using tf
                try:
                    try:
                        from .keras_backend import KerasModelAdapter
                    except Exception:
                        from .tensorflow_backend import KerasModelAdapter

                    return KerasModelAdapter(
                        model_description=model_description, devices=devices
                    )
                except Exception as e:
                    errors.append(e)
            else:
                assert_never(wf)

        assert errors
        if len(weight_format_priority_order) == 1:
            assert len(errors) == 1
            raise errors[0]

        else:
            msg = (
                "None of the weight format specific model adapters could be created"
                + " in this environment."
            )
            raise ExceptionGroup(msg, errors)

    @final
    def load(self, *, devices: Optional[Sequence[str]] = None) -> None:
        warnings.warn("Deprecated. ModelAdapter is loaded on initialization")

    def forward(
        self, input_sample: Union[Sample, SampleBlock, SampleBlockWithOrigin]
    ) -> Sample:
        """
        Run forward pass of model to get model predictions

        Note: sample id and stample stat attributes are passed through
        """
        unexpected = [mid for mid in input_sample.members if mid not in self._input_ids]
        if unexpected:
            warnings.warn(f"Got unexpected input tensor IDs: {unexpected}")

        input_arrays = [
            (
                None
                if (a := input_sample.members.get(in_id)) is None
                else a.transpose(in_order).data.data
            )
            for in_id, in_order in zip(self._input_ids, self._input_axes)
        ]
        output_arrays = self._forward_impl(input_arrays)
        assert len(output_arrays) <= len(self._output_ids)
        output_tensors = [
            None if a is None else Tensor(a, dims=d)
            for a, d in zip(output_arrays, self._output_axes)
        ]
        return Sample(
            members={
                tid: out
                for tid, out in zip(
                    self._output_ids,
                    output_tensors,
                )
                if out is not None
            },
            stat=input_sample.stat,
            id=(
                input_sample.id
                if isinstance(input_sample, Sample)
                else input_sample.sample_id
            ),
        )

    @abstractmethod
    def _forward_impl(
        self, input_arrays: Sequence[Optional[NDArray[Any]]]
    ) -> Union[List[Optional[NDArray[Any]]], Tuple[Optional[NDArray[Any]]]]:
        """framework specific forward implementation"""

    @abstractmethod
    def unload(self):
        """
        Unload model from any devices, freeing their memory.
        The moder adapter should be considered unusable afterwards.
        """

    def _get_input_args_numpy(self, input_sample: Sample):
        """helper to extract tensor args as transposed numpy arrays"""


create_model_adapter = ModelAdapter.create



"""use these type guards with caution!
They widen the type to T[Any], which is not always correct."""

from bioimageio.spec._internal import type_guards

is_list = type_guards.is_list
is_ndarray = type_guards.is_ndarray
is_tuple = type_guards.is_tuple



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Any,
    Dict,
    Literal,
    Mapping,
    Optional,
    Protocol,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
from pydantic import (
    BaseModel,
    BeforeValidator,
    Discriminator,
    PlainSerializer,
)
from typing_extensions import Annotated

from .axis import AxisId
from .common import MemberId, PerMember
from .tensor import Tensor


def tensor_custom_before_validator(data: Union[Tensor, Mapping[str, Any]]):
    if isinstance(data, Tensor):
        return data

    # custom before validation logic
    return Tensor(np.asarray(data["data"]), dims=data["dims"])


def tensor_custom_serializer(t: Tensor) -> Dict[str, Any]:
    # custome serialization logic
    return {"data": t.data.data.tolist(), "dims": list(map(str, t.dims))}


MeasureValue = Union[
    float,
    Annotated[
        Tensor,
        BeforeValidator(tensor_custom_before_validator),
        PlainSerializer(tensor_custom_serializer),
    ],
]


# using Sample Protocol really only to avoid circular imports
class SampleLike(Protocol):
    @property
    def members(self) -> PerMember[Tensor]: ...


class MeasureBase(BaseModel, frozen=True):
    member_id: MemberId


class SampleMeasureBase(MeasureBase, ABC, frozen=True):
    scope: Literal["sample"] = "sample"

    @abstractmethod
    def compute(self, sample: SampleLike) -> MeasureValue:
        """compute the measure"""
        ...


class DatasetMeasureBase(MeasureBase, ABC, frozen=True):
    scope: Literal["dataset"] = "dataset"


class _Mean(BaseModel, frozen=True):
    name: Literal["mean"] = "mean"
    axes: Optional[Tuple[AxisId, ...]] = None
    """`axes` to reduce"""


class SampleMean(_Mean, SampleMeasureBase, frozen=True):
    """The mean value of a single tensor"""

    def compute(self, sample: SampleLike) -> MeasureValue:
        tensor = sample.members[self.member_id]
        return tensor.mean(dim=self.axes)

    def model_post_init(self, __context: Any):
        assert self.axes is None or AxisId("batch") not in self.axes


class DatasetMean(_Mean, DatasetMeasureBase, frozen=True):
    """The mean value across multiple samples"""

    def model_post_init(self, __context: Any):
        assert self.axes is None or AxisId("batch") in self.axes


class _Std(BaseModel, frozen=True):
    name: Literal["std"] = "std"
    axes: Optional[Tuple[AxisId, ...]] = None
    """`axes` to reduce"""


class SampleStd(_Std, SampleMeasureBase, frozen=True):
    """The standard deviation of a single tensor"""

    def compute(self, sample: SampleLike) -> MeasureValue:
        tensor = sample.members[self.member_id]
        return tensor.std(dim=self.axes)

    def model_post_init(self, __context: Any):
        assert self.axes is None or AxisId("batch") not in self.axes


class DatasetStd(_Std, DatasetMeasureBase, frozen=True):
    """The standard deviation across multiple samples"""

    def model_post_init(self, __context: Any):
        assert self.axes is None or AxisId("batch") in self.axes


class _Var(BaseModel, frozen=True):
    name: Literal["var"] = "var"
    axes: Optional[Tuple[AxisId, ...]] = None
    """`axes` to reduce"""


class SampleVar(_Var, SampleMeasureBase, frozen=True):
    """The variance of a single tensor"""

    def compute(self, sample: SampleLike) -> MeasureValue:
        tensor = sample.members[self.member_id]
        return tensor.var(dim=self.axes)

    def model_post_init(self, __context: Any):
        assert self.axes is None or AxisId("batch") not in self.axes


class DatasetVar(_Var, DatasetMeasureBase, frozen=True):
    """The variance across multiple samples"""

    def model_post_init(self, __context: Any):  # TODO: turn into @model_validator
        assert self.axes is None or AxisId("batch") in self.axes


class _Quantile(BaseModel, frozen=True):
    name: Literal["quantile"] = "quantile"
    q: float
    axes: Optional[Tuple[AxisId, ...]] = None
    """`axes` to reduce"""

    def model_post_init(self, __context: Any):
        assert self.q >= 0.0
        assert self.q <= 1.0


class SampleQuantile(_Quantile, SampleMeasureBase, frozen=True):
    """The `n`th percentile of a single tensor"""

    def compute(self, sample: SampleLike) -> MeasureValue:
        tensor = sample.members[self.member_id]
        return tensor.quantile(self.q, dim=self.axes)

    def model_post_init(self, __context: Any):
        super().model_post_init(__context)
        assert self.axes is None or AxisId("batch") not in self.axes


class DatasetPercentile(_Quantile, DatasetMeasureBase, frozen=True):
    """The `n`th percentile across multiple samples"""

    def model_post_init(self, __context: Any):
        super().model_post_init(__context)
        assert self.axes is None or AxisId("batch") in self.axes


SampleMeasure = Annotated[
    Union[SampleMean, SampleStd, SampleVar, SampleQuantile], Discriminator("name")
]
DatasetMeasure = Annotated[
    Union[DatasetMean, DatasetStd, DatasetVar, DatasetPercentile], Discriminator("name")
]
Measure = Annotated[Union[SampleMeasure, DatasetMeasure], Discriminator("scope")]
Stat = Dict[Measure, MeasureValue]

MeanMeasure = Union[SampleMean, DatasetMean]
StdMeasure = Union[SampleStd, DatasetStd]
VarMeasure = Union[SampleVar, DatasetVar]
PercentileMeasure = Union[SampleQuantile, DatasetPercentile]
MeanMeasureT = TypeVar("MeanMeasureT", bound=MeanMeasure)
StdMeasureT = TypeVar("StdMeasureT", bound=StdMeasure)
VarMeasureT = TypeVar("VarMeasureT", bound=VarMeasure)
PercentileMeasureT = TypeVar("PercentileMeasureT", bound=PercentileMeasure)



"""DEPRECATED"""

from typing import List

from .backends._model_adapter import (
    DEFAULT_WEIGHT_FORMAT_PRIORITY_ORDER,
    ModelAdapter,
    create_model_adapter,
)

__all__ = [
    "ModelAdapter",
    "create_model_adapter",
    "get_weight_formats",
]


def get_weight_formats() -> List[str]:
    """
    Return list of supported weight types
    """
    return list(DEFAULT_WEIGHT_FORMAT_PRIORITY_ORDER)



from pathlib import Path
from typing import Tuple

import numpy as np
import pytest


@pytest.mark.parametrize(
    "name",
    [
        "img.png",
        "img.tiff",
        "img.h5",
        "img.h5/img",
        "img.npy",
    ],
)
@pytest.mark.parametrize(
    "shape",
    [
        (4, 5),
        (3, 4, 5),
        (1, 4, 5),
        (5, 4, 3),
        (5, 3, 4),
    ],
)
def test_image_io(name: str, shape: Tuple[int, ...], tmp_path: Path):
    from bioimageio.core import Tensor
    from bioimageio.core.io import load_tensor, save_tensor

    path = tmp_path / name
    data = Tensor.from_numpy(
        np.arange(np.prod(shape), dtype=np.uint8).reshape(shape), dims=None
    )
    save_tensor(path, data)
    actual = load_tensor(path)
    assert actual == data



from typing import Sequence

import numpy as np
import pytest
import xarray as xr
from xarray.testing import assert_equal  # pyright: ignore[reportUnknownVariableType]

from bioimageio.core import AxisId, Tensor


@pytest.mark.parametrize(
    "axes",
    [
        "yx",
        "xy",
        "cyx",
        "yxc",
        "bczyx",
        "xyz",
        "xyzc",
        "bzyxc",
        ("batch", "channel", "x", "y"),
    ],
)
def test_transpose_tensor_2d(axes: Sequence[str]):

    tensor = Tensor.from_numpy(np.random.rand(256, 256), dims=None)
    transposed = tensor.transpose([AxisId(a) for a in axes])
    assert transposed.ndim == len(axes)


@pytest.mark.parametrize(
    "axes",
    [
        "zyx",
        "cyzx",
        "yzixc",
        "bczyx",
        "xyz",
        "xyzc",
        "bzyxtc",
        ("batch", "channel", "x", "y", "z"),
    ],
)
def test_transpose_tensor_3d(axes: Sequence[str]):
    tensor = Tensor.from_numpy(np.random.rand(64, 64, 64), dims=None)
    transposed = tensor.transpose([AxisId(a) for a in axes])
    assert transposed.ndim == len(axes)


def test_crop_and_pad():
    tensor = Tensor.from_xarray(
        xr.DataArray(np.random.rand(10, 20), dims=("x", "y"), name="id")
    )
    padded = tensor.pad({AxisId("x"): 7, AxisId("y"): (3, 3)})
    cropped = padded.crop_to(tensor.sizes)
    assert_equal(tensor.data, cropped.data)


def test_some_magic_ops():
    tensor = Tensor.from_numpy(np.random.rand(256, 256), dims=None)
    assert tensor + 2 == 2 + tensor


def test_shape_attributes():
    tensor = Tensor.from_numpy(np.random.rand(1, 2, 25, 26), dims=None)
    assert tensor.shape_tuple == tensor.shape



# type: ignore  # TODO enable type checking
import os
import zipfile
from pathlib import Path

import pytest

from bioimageio.spec import load_description
from bioimageio.spec.model import v0_5


def test_pytorch_to_torchscript(any_torch_model, tmp_path):
    from bioimageio.core import test_model
    from bioimageio.core.weight_converters.pytorch_to_torchscript import convert

    model_descr = load_description(any_torch_model, perform_io_checks=False)
    if model_descr.implemented_format_version_tuple[:2] == (0, 4):
        pytest.skip("cannot convert to old 0.4 format")

    out_path = tmp_path / "weights.pt"
    ret_val = convert(model_descr, out_path)
    assert out_path.exists()
    assert isinstance(ret_val, v0_5.TorchscriptWeightsDescr)
    assert ret_val.source == out_path
    model_descr.weights.torchscript = ret_val
    summary = test_model(model_descr, weight_format="torchscript")
    assert summary.status == "passed", summary.display()


def test_pytorch_to_onnx(convert_to_onnx, tmp_path):
    from bioimageio.core import test_model
    from bioimageio.core.weight_converters.pytorch_to_onnx import convert

    model_descr = load_description(convert_to_onnx, format_version="latest")
    out_path = tmp_path / "weights.onnx"
    opset_version = 15
    ret_val = convert(
        model_descr=model_descr,
        output_path=out_path,
        opset_version=opset_version,
    )
    assert os.path.exists(out_path)
    assert isinstance(ret_val, v0_5.OnnxWeightsDescr)
    assert ret_val.opset_version == opset_version
    assert ret_val.source == out_path

    model_descr.weights.onnx = ret_val
    summary = test_model(model_descr, weight_format="onnx")
    assert summary.status == "passed", summary.display()


@pytest.mark.skip()
def test_keras_to_tensorflow(any_keras_model: Path, tmp_path: Path):
    from bioimageio.core import test_model
    from bioimageio.core.weight_converters.keras_to_tensorflow import convert

    out_path = tmp_path / "weights.zip"
    model_descr = load_description(any_keras_model)
    ret_val = convert(model_descr, out_path)

    assert out_path.exists()
    assert isinstance(ret_val, v0_5.TensorflowSavedModelBundleWeightsDescr)

    expected_names = {"saved_model.pb", "variables/variables.index"}
    with zipfile.ZipFile(out_path, "r") as f:
        names = set([name for name in f.namelist()])
    assert len(expected_names - names) == 0

    model_descr.weights.keras = ret_val
    summary = test_model(model_descr, weight_format="keras_hdf5")
    assert summary.status == "passed", summary.display()


# TODO: add tensorflow_to_keras converter
# def test_tensorflow_to_keras(any_tensorflow_model: Path, tmp_path: Path):
#     from bioimageio.core.weight_converters.tensorflow_to_keras import convert

#     model_descr = load_description(any_tensorflow_model)
#     out_path = tmp_path / "weights.h5"
#     ret_val = convert(model_descr, output_path=out_path)
#     assert out_path.exists()
#     assert isinstance(ret_val, v0_5.TensorflowSavedModelBundleWeightsDescr)
#     assert ret_val.source == out_path

#     model_descr.weights.keras = ret_val
#     summary = test_model(model_descr, weight_format="keras_hdf5")
#     assert summary.status == "passed", summary.display()


# @pytest.mark.skip()
# def test_tensorflow_to_keras_zipped(any_tensorflow_model: Path, tmp_path: Path):
#     from bioimageio.core.weight_converters.tensorflow_to_keras import convert

#     out_path = tmp_path / "weights.zip"
#     model_descr = load_description(any_tensorflow_model)
#     ret_val = convert(model_descr, out_path)

#     assert out_path.exists()
#     assert isinstance(ret_val, v0_5.TensorflowSavedModelBundleWeightsDescr)

#     expected_names = {"saved_model.pb", "variables/variables.index"}
#     with zipfile.ZipFile(out_path, "r") as f:
#         names = set([name for name in f.namelist()])
#     assert len(expected_names - names) == 0

#     model_descr.weights.keras = ret_val
#     summary = test_model(model_descr, weight_format="keras_hdf5")
#     assert summary.status == "passed", summary.display()



import warnings
from types import MappingProxyType
from typing import (
    Any,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from loguru import logger
from tqdm import tqdm

from bioimageio.spec.model import AnyModelDescr, v0_4, v0_5

from ._op_base import BlockedOperator
from .axis import AxisId, PerAxis
from .common import (
    BlocksizeParameter,
    Halo,
    MemberId,
    PerMember,
    SampleId,
    SupportedWeightsFormat,
)
from .digest_spec import (
    get_block_transform,
    get_input_halo,
    get_member_ids,
)
from .model_adapters import ModelAdapter, create_model_adapter
from .model_adapters import get_weight_formats as get_weight_formats
from .proc_ops import Processing
from .proc_setup import setup_pre_and_postprocessing
from .sample import Sample, SampleBlock, SampleBlockWithOrigin
from .stat_measures import DatasetMeasure, MeasureValue, Stat
from .tensor import Tensor

Predict_IO = TypeVar(
    "Predict_IO",
    Sample,
    Iterable[Sample],
)


class PredictionPipeline:
    """
    Represents model computation including preprocessing and postprocessing
    Note: Ideally use the `PredictionPipeline` in a with statement
        (as a context manager).
    """

    def __init__(
        self,
        *,
        name: str,
        model_description: AnyModelDescr,
        preprocessing: List[Processing],
        postprocessing: List[Processing],
        model_adapter: ModelAdapter,
        default_ns: Optional[BlocksizeParameter] = None,
        default_blocksize_parameter: BlocksizeParameter = 10,
        default_batch_size: int = 1,
    ) -> None:
        """Use `create_prediction_pipeline` to create a `PredictionPipeline`"""
        super().__init__()
        default_blocksize_parameter = default_ns or default_blocksize_parameter
        if default_ns is not None:
            warnings.warn(
                "Argument `default_ns` is deprecated in favor of"
                + " `default_blocksize_paramter` and will be removed soon."
            )
        del default_ns

        if model_description.run_mode:
            warnings.warn(
                f"Not yet implemented inference for run mode '{model_description.run_mode.name}'"
            )

        self.name = name
        self._preprocessing = preprocessing
        self._postprocessing = postprocessing

        self.model_description = model_description
        if isinstance(model_description, v0_4.ModelDescr):
            self._default_input_halo: PerMember[PerAxis[Halo]] = {}
            self._block_transform = None
        else:
            default_output_halo = {
                t.id: {
                    a.id: Halo(a.halo, a.halo)
                    for a in t.axes
                    if isinstance(a, v0_5.WithHalo)
                }
                for t in model_description.outputs
            }
            self._default_input_halo = get_input_halo(
                model_description, default_output_halo
            )
            self._block_transform = get_block_transform(model_description)

        self._default_blocksize_parameter = default_blocksize_parameter
        self._default_batch_size = default_batch_size

        self._input_ids = get_member_ids(model_description.inputs)
        self._output_ids = get_member_ids(model_description.outputs)

        self._adapter: ModelAdapter = model_adapter

    def __enter__(self):
        self.load()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        self.unload()
        return False

    def predict_sample_block(
        self,
        sample_block: SampleBlockWithOrigin,
        skip_preprocessing: bool = False,
        skip_postprocessing: bool = False,
    ) -> SampleBlock:
        if isinstance(self.model_description, v0_4.ModelDescr):
            raise NotImplementedError(
                f"predict_sample_block not implemented for model {self.model_description.format_version}"
            )
        else:
            assert self._block_transform is not None

        if not skip_preprocessing:
            self.apply_preprocessing(sample_block)

        output_meta = sample_block.get_transformed_meta(self._block_transform)
        local_output = self._adapter.forward(sample_block)

        output = output_meta.with_data(local_output.members, stat=local_output.stat)
        if not skip_postprocessing:
            self.apply_postprocessing(output)

        return output

    def predict_sample_without_blocking(
        self,
        sample: Sample,
        skip_preprocessing: bool = False,
        skip_postprocessing: bool = False,
    ) -> Sample:
        """predict a sample.
        The sample's tensor shapes have to match the model's input tensor description.
        If that is not the case, consider `predict_sample_with_blocking`"""

        if not skip_preprocessing:
            self.apply_preprocessing(sample)

        output = self._adapter.forward(sample)
        if not skip_postprocessing:
            self.apply_postprocessing(output)

        return output

    def get_output_sample_id(self, input_sample_id: SampleId):
        warnings.warn(
            "`PredictionPipeline.get_output_sample_id()` is deprecated and will be"
            + " removed soon. Output sample id is equal to input sample id, hence this"
            + " function is not needed."
        )
        return input_sample_id

    def predict_sample_with_fixed_blocking(
        self,
        sample: Sample,
        input_block_shape: Mapping[MemberId, Mapping[AxisId, int]],
        *,
        skip_preprocessing: bool = False,
        skip_postprocessing: bool = False,
    ) -> Sample:
        if not skip_preprocessing:
            self.apply_preprocessing(sample)

        n_blocks, input_blocks = sample.split_into_blocks(
            input_block_shape,
            halo=self._default_input_halo,
            pad_mode="reflect",
        )
        input_blocks = list(input_blocks)
        predicted_blocks: List[SampleBlock] = []
        logger.info(
            "split sample shape {} into {} blocks of {}.",
            {k: dict(v) for k, v in sample.shape.items()},
            n_blocks,
            {k: dict(v) for k, v in input_block_shape.items()},
        )
        for b in tqdm(
            input_blocks,
            desc=f"predict {sample.id or ''} with {self.model_description.id or self.model_description.name}",
            unit="block",
            unit_divisor=1,
            total=n_blocks,
        ):
            predicted_blocks.append(
                self.predict_sample_block(
                    b, skip_preprocessing=True, skip_postprocessing=True
                )
            )

        predicted_sample = Sample.from_blocks(predicted_blocks)
        if not skip_postprocessing:
            self.apply_postprocessing(predicted_sample)

        return predicted_sample

    def predict_sample_with_blocking(
        self,
        sample: Sample,
        skip_preprocessing: bool = False,
        skip_postprocessing: bool = False,
        ns: Optional[
            Union[
                v0_5.ParameterizedSize_N,
                Mapping[Tuple[MemberId, AxisId], v0_5.ParameterizedSize_N],
            ]
        ] = None,
        batch_size: Optional[int] = None,
    ) -> Sample:
        """predict a sample by splitting it into blocks according to the model and the `ns` parameter"""

        if isinstance(self.model_description, v0_4.ModelDescr):
            raise NotImplementedError(
                "`predict_sample_with_blocking` not implemented for v0_4.ModelDescr"
                + f" {self.model_description.name}."
                + " Consider using `predict_sample_with_fixed_blocking`"
            )

        ns = ns or self._default_blocksize_parameter
        if isinstance(ns, int):
            ns = {
                (ipt.id, a.id): ns
                for ipt in self.model_description.inputs
                for a in ipt.axes
                if isinstance(a.size, v0_5.ParameterizedSize)
            }
        input_block_shape = self.model_description.get_tensor_sizes(
            ns, batch_size or self._default_batch_size
        ).inputs

        return self.predict_sample_with_fixed_blocking(
            sample,
            input_block_shape=input_block_shape,
            skip_preprocessing=skip_preprocessing,
            skip_postprocessing=skip_postprocessing,
        )

    # def predict(
    #     self,
    #     inputs: Predict_IO,
    #     skip_preprocessing: bool = False,
    #     skip_postprocessing: bool = False,
    # ) -> Predict_IO:
    #     """Run model prediction **including** pre/postprocessing."""

    #     if isinstance(inputs, Sample):
    #         return self.predict_sample_with_blocking(
    #             inputs,
    #             skip_preprocessing=skip_preprocessing,
    #             skip_postprocessing=skip_postprocessing,
    #         )
    #     elif isinstance(inputs, collections.abc.Iterable):
    #         return (
    #             self.predict(
    #                 ipt,
    #                 skip_preprocessing=skip_preprocessing,
    #                 skip_postprocessing=skip_postprocessing,
    #             )
    #             for ipt in inputs
    #         )
    #     else:
    #         assert_never(inputs)

    def apply_preprocessing(self, sample: Union[Sample, SampleBlockWithOrigin]) -> None:
        """apply preprocessing in-place, also updates sample stats"""
        for op in self._preprocessing:
            op(sample)

    def apply_postprocessing(
        self, sample: Union[Sample, SampleBlock, SampleBlockWithOrigin]
    ) -> None:
        """apply postprocessing in-place, also updates samples stats"""
        for op in self._postprocessing:
            if isinstance(sample, (Sample, SampleBlockWithOrigin)):
                op(sample)
            elif not isinstance(op, BlockedOperator):
                raise NotImplementedError(
                    "block wise update of output statistics not yet implemented"
                )
            else:
                op(sample)

    def load(self):
        """
        optional step: load model onto devices before calling forward if not using it as context manager
        """
        pass

    def unload(self):
        """
        free any device memory in use
        """
        self._adapter.unload()


def create_prediction_pipeline(
    bioimageio_model: AnyModelDescr,
    *,
    devices: Optional[Sequence[str]] = None,
    weight_format: Optional[SupportedWeightsFormat] = None,
    weights_format: Optional[SupportedWeightsFormat] = None,
    dataset_for_initial_statistics: Iterable[Union[Sample, Sequence[Tensor]]] = tuple(),
    keep_updating_initial_dataset_statistics: bool = False,
    fixed_dataset_statistics: Mapping[DatasetMeasure, MeasureValue] = MappingProxyType(
        {}
    ),
    model_adapter: Optional[ModelAdapter] = None,
    ns: Optional[BlocksizeParameter] = None,
    default_blocksize_parameter: BlocksizeParameter = 10,
    **deprecated_kwargs: Any,
) -> PredictionPipeline:
    """
    Creates prediction pipeline which includes:
    * computation of input statistics
    * preprocessing
    * model prediction
    * computation of output statistics
    * postprocessing

    Args:
        bioimageio_model: A bioimageio model description.
        devices: (optional)
        weight_format: deprecated in favor of **weights_format**
        weights_format: (optional) Use a specific **weights_format** rather than
            choosing one automatically.
            A corresponding `bioimageio.core.model_adapters.ModelAdapter` will be
            created to run inference with the **bioimageio_model**.
        dataset_for_initial_statistics: (optional) If preprocessing steps require input
            dataset statistics, **dataset_for_initial_statistics** allows you to
            specifcy a dataset from which these statistics are computed.
        keep_updating_initial_dataset_statistics: (optional) Set to `True` if you want
            to update dataset statistics with each processed sample.
        fixed_dataset_statistics: (optional) Allows you to specify a mapping of
            `DatasetMeasure`s to precomputed `MeasureValue`s.
        model_adapter: (optional) Allows you to use a custom **model_adapter** instead
            of creating one according to the present/selected **weights_format**.
        ns: deprecated in favor of **default_blocksize_parameter**
        default_blocksize_parameter: Allows to control the default block size for
            blockwise predictions, see `BlocksizeParameter`.

    """
    weights_format = weight_format or weights_format
    del weight_format
    default_blocksize_parameter = ns or default_blocksize_parameter
    del ns
    if deprecated_kwargs:
        warnings.warn(
            f"deprecated create_prediction_pipeline kwargs: {set(deprecated_kwargs)}"
        )

    model_adapter = model_adapter or create_model_adapter(
        model_description=bioimageio_model,
        devices=devices,
        weight_format_priority_order=weights_format and (weights_format,),
    )

    input_ids = get_member_ids(bioimageio_model.inputs)

    def dataset():
        common_stat: Stat = {}
        for i, x in enumerate(dataset_for_initial_statistics):
            if isinstance(x, Sample):
                yield x
            else:
                yield Sample(members=dict(zip(input_ids, x)), stat=common_stat, id=i)

    preprocessing, postprocessing = setup_pre_and_postprocessing(
        bioimageio_model,
        dataset(),
        keep_updating_initial_dataset_stats=keep_updating_initial_dataset_statistics,
        fixed_dataset_stats=fixed_dataset_statistics,
    )

    return PredictionPipeline(
        name=bioimageio_model.name,
        model_description=bioimageio_model,
        model_adapter=model_adapter,
        preprocessing=preprocessing,
        postprocessing=postprocessing,
        default_blocksize_parameter=default_blocksize_parameter,
    )



from collections import defaultdict
from itertools import chain
from typing import DefaultDict, Dict

from bioimageio.spec.model.v0_5 import ModelDescr


def get_dynamic_axes(model_descr: ModelDescr):
    dynamic_axes: DefaultDict[str, Dict[int, str]] = defaultdict(dict)
    for d in chain(model_descr.inputs, model_descr.outputs):
        for i, ax in enumerate(d.axes):
            if not isinstance(ax.size, int):
                dynamic_axes[str(d.id)][i] = str(ax.id)

    return dynamic_axes



import gc
import warnings
from contextlib import nullcontext
from io import BytesIO, TextIOWrapper
from pathlib import Path
from typing import Any, List, Literal, Optional, Sequence, Union

import torch
from loguru import logger
from numpy.typing import NDArray
from torch import nn
from typing_extensions import assert_never

from bioimageio.spec._internal.version_type import Version
from bioimageio.spec.common import BytesReader, ZipPath
from bioimageio.spec.model import AnyModelDescr, v0_4, v0_5
from bioimageio.spec.utils import download

from ..digest_spec import import_callable
from ..utils._type_guards import is_list, is_ndarray, is_tuple
from ._model_adapter import ModelAdapter


class PytorchModelAdapter(ModelAdapter):
    def __init__(
        self,
        *,
        model_description: AnyModelDescr,
        devices: Optional[Sequence[Union[str, torch.device]]] = None,
        mode: Literal["eval", "train"] = "eval",
    ):
        super().__init__(model_description=model_description)
        weights = model_description.weights.pytorch_state_dict
        if weights is None:
            raise ValueError("No `pytorch_state_dict` weights found")

        devices = get_devices(devices)
        self._model = load_torch_model(weights, load_state=True, devices=devices)
        if mode == "eval":
            self._model = self._model.eval()
        elif mode == "train":
            self._model = self._model.train()
        else:
            assert_never(mode)

        self._mode: Literal["eval", "train"] = mode
        self._primary_device = devices[0]

    def _forward_impl(
        self, input_arrays: Sequence[Optional[NDArray[Any]]]
    ) -> List[Optional[NDArray[Any]]]:
        tensors = [
            None if a is None else torch.from_numpy(a).to(self._primary_device)
            for a in input_arrays
        ]

        if self._mode == "eval":
            ctxt = torch.no_grad
        elif self._mode == "train":
            ctxt = nullcontext
        else:
            assert_never(self._mode)

        with ctxt():
            model_out = self._model(*tensors)

        if is_tuple(model_out) or is_list(model_out):
            model_out_seq = model_out
        else:
            model_out_seq = model_out = [model_out]

        result: List[Optional[NDArray[Any]]] = []
        for i, r in enumerate(model_out_seq):
            if r is None:
                result.append(None)
            elif isinstance(r, torch.Tensor):
                r_np: NDArray[Any] = (  # pyright: ignore[reportUnknownVariableType]
                    r.detach().cpu().numpy()
                )
                result.append(r_np)
            elif is_ndarray(r):
                result.append(r)
            else:
                raise TypeError(f"Model output[{i}] has unexpected type {type(r)}.")

        return result

    def unload(self) -> None:
        del self._model
        _ = gc.collect()  # deallocate memory
        assert torch is not None
        torch.cuda.empty_cache()  # release reserved memory


def load_torch_model(
    weight_spec: Union[
        v0_4.PytorchStateDictWeightsDescr, v0_5.PytorchStateDictWeightsDescr
    ],
    *,
    load_state: bool = True,
    devices: Optional[Sequence[Union[str, torch.device]]] = None,
) -> nn.Module:
    custom_callable = import_callable(
        weight_spec.architecture,
        sha256=(
            weight_spec.architecture_sha256
            if isinstance(weight_spec, v0_4.PytorchStateDictWeightsDescr)
            else weight_spec.sha256
        ),
    )
    model_kwargs = (
        weight_spec.kwargs
        if isinstance(weight_spec, v0_4.PytorchStateDictWeightsDescr)
        else weight_spec.architecture.kwargs
    )
    torch_model = custom_callable(**model_kwargs)

    if not isinstance(torch_model, nn.Module):
        if isinstance(
            weight_spec.architecture,
            (v0_4.CallableFromFile, v0_4.CallableFromDepencency),
        ):
            callable_name = weight_spec.architecture.callable_name
        else:
            callable_name = weight_spec.architecture.callable

        raise ValueError(f"Calling {callable_name} did not return a torch.nn.Module.")

    if load_state or devices:
        use_devices = get_devices(devices)
        torch_model = torch_model.to(use_devices[0])
        if load_state:
            torch_model = load_torch_state_dict(
                torch_model,
                path=download(weight_spec),
                devices=use_devices,
            )
    return torch_model


def load_torch_state_dict(
    model: nn.Module,
    path: Union[Path, ZipPath, BytesReader],
    devices: Sequence[torch.device],
) -> nn.Module:
    model = model.to(devices[0])
    if isinstance(path, (Path, ZipPath)):
        ctxt = path.open("rb")
    else:
        ctxt = nullcontext(BytesIO(path.read()))

    with ctxt as f:
        assert not isinstance(f, TextIOWrapper)
        if Version(str(torch.__version__)) < Version("1.13"):
            state = torch.load(f, map_location=devices[0])
        else:
            state = torch.load(f, map_location=devices[0], weights_only=True)

    incompatible = model.load_state_dict(state)
    if (
        isinstance(incompatible, tuple)
        and hasattr(incompatible, "missing_keys")
        and hasattr(incompatible, "unexpected_keys")
    ):
        if incompatible.missing_keys:
            logger.warning("Missing state dict keys: {}", incompatible.missing_keys)

        if hasattr(incompatible, "unexpected_keys") and incompatible.unexpected_keys:
            logger.warning(
                "Unexpected state dict keys: {}", incompatible.unexpected_keys
            )
    else:
        logger.warning(
            "`model.load_state_dict()` unexpectedly returned: {} "
            + "(expected named tuple with `missing_keys` and `unexpected_keys` attributes)",
            (s[:20] + "..." if len(s := str(incompatible)) > 20 else s),
        )

    return model


def get_devices(
    devices: Optional[Sequence[Union[torch.device, str]]] = None,
) -> List[torch.device]:
    if not devices:
        torch_devices = [
            torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        ]
    else:
        torch_devices = [torch.device(d) for d in devices]

    if len(torch_devices) > 1:
        warnings.warn(
            f"Multiple devices for single pytorch model not yet implemented; ignoring {torch_devices[1:]}"
        )
        torch_devices = torch_devices[:1]

    return torch_devices



import traceback
from typing import Optional, Union

from loguru import logger
from pydantic import DirectoryPath

from bioimageio.spec import (
    InvalidDescr,
    load_model_description,
    save_bioimageio_package_as_folder,
)
from bioimageio.spec.model.v0_5 import ModelDescr, WeightsFormat

from .._resource_tests import load_description_and_test


def add_weights(
    model_descr: ModelDescr,
    *,
    output_path: DirectoryPath,
    source_format: Optional[WeightsFormat] = None,
    target_format: Optional[WeightsFormat] = None,
    verbose: bool = False,
    allow_tracing: bool = True,
) -> Union[ModelDescr, InvalidDescr]:
    """Convert model weights to other formats and add them to the model description

    Args:
        output_path: Path to save updated model package to.
        source_format: convert from a specific weights format.
                       Default: choose automatically from any available.
        target_format: convert to a specific weights format.
                       Default: attempt to convert to any missing format.
        devices: Devices that may be used during conversion.
        verbose: log more (error) output

    Returns:
        A (potentially invalid) model copy stored at `output_path` with added weights if any conversion was possible.

    """
    if not isinstance(model_descr, ModelDescr):
        if model_descr.type == "model" and not isinstance(model_descr, InvalidDescr):
            raise TypeError(
                f"Model format {model_descr.format} is not supported, please update"
                + f" model to format {ModelDescr.implemented_format_version} first."
            )

        raise TypeError(type(model_descr))

    # save model to local folder
    output_path = save_bioimageio_package_as_folder(
        model_descr, output_path=output_path
    )
    # reload from local folder to make sure we do not edit the given model
    model_descr = load_model_description(
        output_path, perform_io_checks=False, format_version="latest"
    )

    if source_format is None:
        available = set(model_descr.weights.available_formats)
    else:
        available = {source_format}

    if target_format is None:
        missing = set(model_descr.weights.missing_formats)
    else:
        missing = {target_format}

    originally_missing = set(missing)

    if "pytorch_state_dict" in available and "torchscript" in missing:
        logger.info(
            "Attempting to convert 'pytorch_state_dict' weights to 'torchscript'."
        )
        from .pytorch_to_torchscript import convert

        try:
            torchscript_weights_path = output_path / "weights_torchscript.pt"
            model_descr.weights.torchscript = convert(
                model_descr,
                output_path=torchscript_weights_path,
                use_tracing=False,
            )
        except Exception as e:
            if verbose:
                traceback.print_exception(type(e), e, e.__traceback__)

            logger.error(e)
        else:
            available.add("torchscript")
            missing.discard("torchscript")

    if allow_tracing and "pytorch_state_dict" in available and "torchscript" in missing:
        logger.info(
            "Attempting to convert 'pytorch_state_dict' weights to 'torchscript' by tracing."
        )
        from .pytorch_to_torchscript import convert

        try:
            torchscript_weights_path = output_path / "weights_torchscript_traced.pt"

            model_descr.weights.torchscript = convert(
                model_descr,
                output_path=torchscript_weights_path,
                use_tracing=True,
            )
        except Exception as e:
            if verbose:
                traceback.print_exception(type(e), e, e.__traceback__)

            logger.error(e)
        else:
            available.add("torchscript")
            missing.discard("torchscript")

    if "torchscript" in available and "onnx" in missing:
        logger.info("Attempting to convert 'torchscript' weights to 'onnx'.")
        from .torchscript_to_onnx import convert

        try:
            onnx_weights_path = output_path / "weights.onnx"
            model_descr.weights.onnx = convert(
                model_descr,
                output_path=onnx_weights_path,
            )
        except Exception as e:
            if verbose:
                traceback.print_exception(type(e), e, e.__traceback__)

            logger.error(e)
        else:
            available.add("onnx")
            missing.discard("onnx")

    if "pytorch_state_dict" in available and "onnx" in missing:
        logger.info("Attempting to convert 'pytorch_state_dict' weights to 'onnx'.")
        from .pytorch_to_onnx import convert

        try:
            onnx_weights_path = output_path / "weights.onnx"

            model_descr.weights.onnx = convert(
                model_descr,
                output_path=onnx_weights_path,
                verbose=verbose,
            )
        except Exception as e:
            if verbose:
                traceback.print_exception(type(e), e, e.__traceback__)

            logger.error(e)
        else:
            available.add("onnx")
            missing.discard("onnx")

    if missing:
        logger.warning(
            f"Converting from any of the available weights formats {available} to any"
            + f" of {missing} failed or is not yet implemented. Please create an issue"
            + " at https://github.com/bioimage-io/core-bioimage-io-python/issues/new/choose"
            + " if you would like bioimageio.core to support a particular conversion."
        )

    if originally_missing == missing:
        logger.warning("failed to add any converted weights")
        return model_descr
    else:
        logger.info("added weights formats {}", originally_missing - missing)
        # resave model with updated rdf.yaml
        _ = save_bioimageio_package_as_folder(model_descr, output_path=output_path)
        tested_model_descr = load_description_and_test(
            model_descr, format_version="latest", expected_type="model"
        )
        if not isinstance(tested_model_descr, ModelDescr):
            logger.error(
                f"The updated model description at {output_path} did not pass testing."
            )

        return tested_model_descr



import subprocess
from pathlib import Path
from typing import Any, List, Sequence

import pytest
from pydantic import FilePath


def run_subprocess(
    commands: Sequence[str], **kwargs: Any
) -> "subprocess.CompletedProcess[str]":
    return subprocess.run(
        commands,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        encoding="utf-8",
        **kwargs,
    )


@pytest.mark.parametrize(
    "args",
    [
        [
            "package",
            "unet2d_nuclei_broad_model",
            "output.zip",
            "--weight_format",
            "pytorch_state_dict",
        ],
        ["package", "unet2d_nuclei_broad_model", "output.zip"],
        [
            "test",
            "unet2d_nuclei_broad_model",
            "--weight_format",
            "pytorch_state_dict",
        ],
        ["test", "unet2d_nuclei_broad_model"],
        ["predict", "--example", "unet2d_nuclei_broad_model"],
        ["update-format", "unet2d_nuclei_broad_model_old"],
        ["add-weights", "unet2d_nuclei_broad_model", "tmp_path"],
        ["update-hashes", "unet2d_nuclei_broad_model_old"],
        ["update-hashes", "unet2d_nuclei_broad_model_old", "--output=stdout"],
    ],
)
def test_cli(
    args: List[str],
    unet2d_nuclei_broad_model: str,
    unet2d_nuclei_broad_model_old: str,
    tmp_path: Path,
):
    resolved_args = [
        (
            unet2d_nuclei_broad_model
            if arg == "unet2d_nuclei_broad_model"
            else (
                unet2d_nuclei_broad_model_old
                if arg == "unet2d_nuclei_broad_model_old"
                else (
                    arg.replace("tmp_path", str(tmp_path)) if "tmp_path" in arg else arg
                )
            )
        )
        for arg in args
    ]
    ret = run_subprocess(["bioimageio", *resolved_args])
    assert ret.returncode == 0, ret.stdout


@pytest.mark.parametrize("args", [["test", "stardist_wrong_shape"]])
def test_cli_fails(args: List[str], stardist_wrong_shape: FilePath):
    resolved_args = [
        str(stardist_wrong_shape) if arg == "stardist_wrong_shape" else arg
        for arg in args
    ]
    ret = run_subprocess(["bioimageio", *resolved_args])
    assert ret.returncode == 1, ret.stdout


# TODO: update CLI test
# def _test_cli_predict_image(model: Path, tmp_path: Path, extra_cmd_args: Optional[List[str]] = None):
#     spec = load_description(model)
#     in_path = spec.test_inputs[0]

#     out_path = tmp_path.with_suffix(".npy")
#     cmd = ["bioimageio", "predict-image", model, "--input", str(in_path), "--output", str(out_path)]
#     if extra_cmd_args is not None:
#         cmd.extend(extra_cmd_args)
#     ret = run_subprocess(cmd)
#     assert ret.returncode == 0, ret.stdout
#     assert out_path.exists()


# def test_cli_predict_image(unet2d_nuclei_broad_model: Path, tmp_path: Path):
#     _test_cli_predict_image(unet2d_nuclei_broad_model, tmp_path)


# def test_cli_predict_image_with_weight_format(unet2d_nuclei_broad_model: Path, tmp_path: Path):
#     _test_cli_predict_image(unet2d_nuclei_broad_model, tmp_path, ["--weight-format", "pytorch_state_dict"])


# def _test_cli_predict_images(model: Path, tmp_path: Path, extra_cmd_args: Optional[List[str]] = None):
#     n_images = 3
#     shape = (1, 1, 128, 128)
#     expected_shape = (1, 1, 128, 128)

#     in_folder = tmp_path / "inputs"
#     in_folder.mkdir()
#     out_folder = tmp_path / "outputs"
#     out_folder.mkdir()

#     expected_outputs: List[Path] = []
#     for i in range(n_images):
#         path = in_folder / f"im-{i}.npy"
#         im = np.random.randint(0, 255, size=shape).astype("uint8")
#         np.save(path, im)
#         expected_outputs.append(out_folder / f"im-{i}.npy")

#     input_pattern = str(in_folder / "*.npy")
#     cmd = ["bioimageio", "predict-images", str(model), input_pattern, str(out_folder)]
#     if extra_cmd_args is not None:
#         cmd.extend(extra_cmd_args)
#     ret = run_subprocess(cmd)
#     assert ret.returncode == 0, ret.stdout

#     for out_path in expected_outputs:
#         assert out_path.exists()
#         assert np.load(out_path).shape == expected_shape


# def test_cli_predict_images(unet2d_nuclei_broad_model: Path, tmp_path: Path):
#     _test_cli_predict_images(unet2d_nuclei_broad_model, tmp_path)


# def test_cli_predict_images_with_weight_format(unet2d_nuclei_broad_model: Path, tmp_path: Path):
#     _test_cli_predict_images(unet2d_nuclei_broad_model, tmp_path, ["--weight-format", "pytorch_state_dict"])


# def test_torch_to_torchscript(unet2d_nuclei_broad_model: Path, tmp_path: Path):
#     out_path = tmp_path.with_suffix(".pt")
#     ret = run_subprocess(
#         ["bioimageio", "convert-torch-weights-to-torchscript", str(unet2d_nuclei_broad_model), str(out_path)]
#     )
#     assert ret.returncode == 0, ret.stdout
#     assert out_path.exists()


# def test_torch_to_onnx(convert_to_onnx: Path, tmp_path: Path):
#     out_path = tmp_path.with_suffix(".onnx")
#     ret = run_subprocess(["bioimageio", "convert-torch-weights-to-onnx", str(convert_to_onnx), str(out_path)])
#     assert ret.returncode == 0, ret.stdout
#     assert out_path.exists()


# def test_keras_to_tf(unet2d_keras: Path, tmp_path: Path):
#     out_path = tmp_path / "weights.zip"
#     ret = run_subprocess(["bioimageio", "convert-keras-weights-to-tensorflow", str(unet2d_keras), str(out_path)])
#     assert ret.returncode == 0, ret.stdout
#     assert out_path.exists()



from pathlib import Path
from typing import Any, Optional, Sequence, Union

import numpy as np
import tensorflow as tf  # pyright: ignore[reportMissingTypeStubs]
from loguru import logger
from numpy.typing import NDArray

from bioimageio.core.io import ensure_unzipped
from bioimageio.spec.model import AnyModelDescr, v0_4, v0_5

from ._model_adapter import ModelAdapter


class TensorflowModelAdapter(ModelAdapter):
    weight_format = "tensorflow_saved_model_bundle"

    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ):
        super().__init__(model_description=model_description)

        weight_file = model_description.weights.tensorflow_saved_model_bundle
        if model_description.weights.tensorflow_saved_model_bundle is None:
            raise ValueError("No `tensorflow_saved_model_bundle` weights found")

        if devices is not None:
            logger.warning(
                f"Device management is not implemented for tensorflow yet, ignoring the devices {devices}"
            )

        # TODO: check how to load tf weights without unzipping
        weight_file = ensure_unzipped(
            model_description.weights.tensorflow_saved_model_bundle.source,
            Path("bioimageio_unzipped_tf_weights"),
        )
        self._network = str(weight_file)

    # TODO currently we relaod the model every time. it would be better to keep the graph and session
    # alive in between of forward passes (but then the sessions need to be properly opened / closed)
    def _forward_impl(  # pyright: ignore[reportUnknownParameterType]
        self, input_arrays: Sequence[Optional[NDArray[Any]]]
    ):
        # TODO read from spec
        tag = (  # pyright: ignore[reportUnknownVariableType]
            tf.saved_model.tag_constants.SERVING  # pyright: ignore[reportAttributeAccessIssue]
        )
        signature_key = (  # pyright: ignore[reportUnknownVariableType]
            tf.saved_model.signature_constants.DEFAULT_SERVING_SIGNATURE_DEF_KEY  # pyright: ignore[reportAttributeAccessIssue]
        )

        graph = tf.Graph()
        with graph.as_default():
            with tf.Session(  # pyright: ignore[reportAttributeAccessIssue]
                graph=graph
            ) as sess:  # pyright: ignore[reportUnknownVariableType]
                # load the model and the signature
                graph_def = tf.saved_model.loader.load(  # pyright: ignore[reportUnknownVariableType,reportAttributeAccessIssue]
                    sess, [tag], self._network
                )
                signature = (  # pyright: ignore[reportUnknownVariableType]
                    graph_def.signature_def
                )

                # get the tensors into the graph
                in_names = [  # pyright: ignore[reportUnknownVariableType]
                    signature[signature_key].inputs[key].name for key in self._input_ids
                ]
                out_names = [  # pyright: ignore[reportUnknownVariableType]
                    signature[signature_key].outputs[key].name
                    for key in self._output_ids
                ]
                in_tf_tensors = [
                    graph.get_tensor_by_name(
                        name  # pyright: ignore[reportUnknownArgumentType]
                    )
                    for name in in_names  # pyright: ignore[reportUnknownVariableType]
                ]
                out_tf_tensors = [
                    graph.get_tensor_by_name(
                        name  # pyright: ignore[reportUnknownArgumentType]
                    )
                    for name in out_names  # pyright: ignore[reportUnknownVariableType]
                ]

                # run prediction
                res = sess.run(  # pyright: ignore[reportUnknownVariableType]
                    dict(
                        zip(
                            out_names,  # pyright: ignore[reportUnknownArgumentType]
                            out_tf_tensors,
                        )
                    ),
                    dict(zip(in_tf_tensors, input_arrays)),
                )
                # from dict to list of tensors
                res = [  # pyright: ignore[reportUnknownVariableType]
                    res[out]
                    for out in out_names  # pyright: ignore[reportUnknownVariableType]
                ]

        return res  # pyright: ignore[reportUnknownVariableType]

    def unload(self) -> None:
        logger.warning(
            "Device management is not implemented for tensorflow 1, cannot unload model"
        )


class KerasModelAdapter(ModelAdapter):
    def __init__(
        self,
        *,
        model_description: Union[v0_4.ModelDescr, v0_5.ModelDescr],
        devices: Optional[Sequence[str]] = None,
    ):
        if model_description.weights.tensorflow_saved_model_bundle is None:
            raise ValueError("No `tensorflow_saved_model_bundle` weights found")

        super().__init__(model_description=model_description)
        if devices is not None:
            logger.warning(
                f"Device management is not implemented for tensorflow yet, ignoring the devices {devices}"
            )

        # TODO: check how to load tf weights without unzipping
        weight_file = ensure_unzipped(
            model_description.weights.tensorflow_saved_model_bundle.source,
            Path("bioimageio_unzipped_tf_weights"),
        )

        try:
            self._network = tf.keras.layers.TFSMLayer(  # pyright: ignore[reportAttributeAccessIssue]
                weight_file,
                call_endpoint="serve",
            )
        except Exception as e:
            try:
                self._network = tf.keras.layers.TFSMLayer(  # pyright: ignore[reportAttributeAccessIssue]
                    weight_file, call_endpoint="serving_default"
                )
            except Exception as ee:
                logger.opt(exception=ee).info(
                    "keras.layers.TFSMLayer error for alternative call_endpoint='serving_default'"
                )
                raise e

    def _forward_impl(  # pyright: ignore[reportUnknownParameterType]
        self, input_arrays: Sequence[Optional[NDArray[Any]]]
    ):
        assert tf is not None
        tf_tensor = [
            None if ipt is None else tf.convert_to_tensor(ipt) for ipt in input_arrays
        ]

        result = self._network(*tf_tensor)  # pyright: ignore[reportUnknownVariableType]

        assert isinstance(result, dict)

        # TODO: Use RDF's `outputs[i].id` here
        result = list(  # pyright: ignore[reportUnknownVariableType]
            result.values()  # pyright: ignore[reportUnknownArgumentType]
        )

        return [  # pyright: ignore[reportUnknownVariableType]
            (None if r is None else r if isinstance(r, np.ndarray) else r.numpy())
            for r in result  # pyright: ignore[reportUnknownVariableType]
        ]

    def unload(self) -> None:
        logger.warning(
            "Device management is not implemented for tensorflow>=2 models"
            + f" using `{self.__class__.__name__}`, cannot unload model"
        )


def create_tf_model_adapter(
    model_description: AnyModelDescr, devices: Optional[Sequence[str]]
):
    tf_version = v0_5.Version(tf.__version__)
    weights = model_description.weights.tensorflow_saved_model_bundle
    if weights is None:
        raise ValueError("No `tensorflow_saved_model_bundle` weights found")

    model_tf_version = weights.tensorflow_version
    if model_tf_version is None:
        logger.warning(
            "The model does not specify the tensorflow version."
            + f"Cannot check if it is compatible with intalled tensorflow {tf_version}."
        )
    elif model_tf_version > tf_version:
        logger.warning(
            f"The model specifies a newer tensorflow version than installed: {model_tf_version} > {tf_version}."
        )
    elif (model_tf_version.major, model_tf_version.minor) != (
        tf_version.major,
        tf_version.minor,
    ):
        logger.warning(
            "The tensorflow version specified by the model does not match the installed: "
            + f"{model_tf_version} != {tf_version}."
        )

    if tf_version.major <= 1:
        return TensorflowModelAdapter(
            model_description=model_description, devices=devices
        )
    else:
        return KerasModelAdapter(model_description=model_description, devices=devices)



# TODO: update add weights tests
# import os


# def _test_add_weights(model, tmp_path, base_weights, added_weights, **kwargs):
#     from bioimageio.core.build_spec import add_weights

#     rdf = load_raw_resource_description(model)
#     assert base_weights in rdf.weights
#     assert added_weights in rdf.weights

#     weight_path = load_description(model).weights[added_weights].source
#     assert weight_path.exists()

#     drop_weights = set(rdf.weights.keys()) - {base_weights}
#     for drop in drop_weights:
#         rdf.weights.pop(drop)
#     assert tuple(rdf.weights.keys()) == (base_weights,)

#     in_path = tmp_path / "model1.zip"
#     export_resource_package(rdf, output_path=in_path)

#     out_path = tmp_path / "model2.zip"
#     add_weights(in_path, weight_path, weight_type=added_weights, output_path=out_path, **kwargs)

#     assert out_path.exists()
#     new_rdf = load_description(out_path)
#     assert set(new_rdf.weights.keys()) == {base_weights, added_weights}
#     for weight in new_rdf.weights.values():
#         assert weight.source.exists()

#     test_res = _test_model(out_path, added_weights)
#     failed = [s for s in test_res if s["status"] != "passed"]
#     assert not failed, failed
#     test_res = _test_model(out_path)
#     failed = [s for s in test_res if s["status"] != "passed"]
#     assert not failed, failed

#     # make sure the weights were cleaned from the cwd
#     assert not os.path.exists(os.path.split(weight_path)[1])


# def test_add_torchscript(unet2d_nuclei_broad_model, tmp_path):
#     _test_add_weights(unet2d_nuclei_broad_model, tmp_path, "pytorch_state_dict", "torchscript")


# def test_add_onnx(unet2d_nuclei_broad_model, tmp_path):
#     _test_add_weights(unet2d_nuclei_broad_model, tmp_path, "pytorch_state_dict", "onnx", opset_version=12)



from pathlib import Path
from typing import Literal, Mapping, NamedTuple

import numpy as np
import pytest
import xarray as xr
from typing_extensions import assert_never

from bioimageio.core import (
    AxisId,
    MemberId,
    PredictionPipeline,
    Sample,
    create_prediction_pipeline,
    load_model,
    predict,
)
from bioimageio.core.digest_spec import get_test_input_sample, get_test_output_sample
from bioimageio.spec import AnyModelDescr


def _assert_equal_samples(expected: Sample, actual: Sample):
    assert expected.id == actual.id
    assert expected.members == actual.members


class Prep(NamedTuple):
    model: AnyModelDescr
    prediction_pipeline: PredictionPipeline
    input_sample: Sample
    output_sample: Sample


@pytest.fixture(scope="module")
def prep(any_model: str):
    model = load_model(any_model, perform_io_checks=False)
    input_sample = get_test_input_sample(model)
    output_sample = get_test_output_sample(model)
    return Prep(model, create_prediction_pipeline(model), input_sample, output_sample)


def test_predict_with_pipeline(prep: Prep):
    out = predict(
        model=prep.prediction_pipeline,
        inputs=prep.input_sample,
    )
    _assert_equal_samples(out, prep.output_sample)


@pytest.mark.parametrize("tensor_input", ["numpy", "xarray"])
def test_predict_with_model_description(
    tensor_input: Literal["numpy", "xarray"], prep: Prep
):
    if tensor_input == "xarray":
        ipt = {m: t.data for m, t in prep.input_sample.members.items()}
        assert all(isinstance(v, xr.DataArray) for v in ipt.values())
    elif tensor_input == "numpy":
        ipt = {m: t.data.data for m, t in prep.input_sample.members.items()}
        assert all(isinstance(v, np.ndarray) for v in ipt.values())
    else:
        assert_never(tensor_input)

    out = predict(
        model=prep.model,
        inputs=ipt,
        sample_id=prep.input_sample.id,
        skip_preprocessing=False,
        skip_postprocessing=False,
    )
    _assert_equal_samples(out, prep.output_sample)


@pytest.mark.parametrize("with_procs", [True, False])
def test_predict_with_blocking(with_procs: bool, prep: Prep):
    try:
        out = predict(
            model=prep.prediction_pipeline,
            inputs=prep.input_sample,
            blocksize_parameter=3,
            sample_id=prep.input_sample.id,
            skip_preprocessing=with_procs,
            skip_postprocessing=with_procs,
        )
    except NotImplementedError as e:
        pytest.skip(str(e))

    if with_procs:
        _assert_equal_samples(out, prep.output_sample)
    else:
        assert isinstance(out, Sample)


def test_predict_with_fixed_blocking(prep: Prep):
    block_along = list(prep.input_sample.members)
    input_block_shape: Mapping[MemberId, Mapping[AxisId, int]] = {
        ba: {
            "x": min(  # pyright: ignore[reportAssignmentType]
                128, prep.input_sample.members[ba].tagged_shape[AxisId("x")]
            ),
            AxisId("y"): min(
                128, prep.input_sample.members[ba].tagged_shape[AxisId("y")]
            ),
        }
        for ba in block_along
    }
    try:
        out = predict(
            model=prep.prediction_pipeline,
            inputs=prep.input_sample,
            input_block_shape=input_block_shape,
            sample_id=prep.input_sample.id,
        )
    except NotImplementedError as e:
        pytest.skip(str(e))

    _assert_equal_samples(out, prep.output_sample)


def test_predict_save_output(prep: Prep, tmp_path: Path):
    save_path = tmp_path / "{member_id}_{sample_id}.h5"
    out = predict(
        model=prep.prediction_pipeline,
        inputs=prep.input_sample,
        save_output_path=save_path,
    )
    _assert_equal_samples(out, prep.output_sample)
    assert save_path.parent.exists()



from __future__ import annotations

import collections.abc
import importlib.util
import sys
from itertools import chain
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)
from zipfile import ZipFile, is_zipfile

import numpy as np
import xarray as xr
from loguru import logger
from numpy.typing import NDArray
from typing_extensions import Unpack, assert_never

from bioimageio.spec._internal.io import HashKwargs
from bioimageio.spec.common import FileDescr, FileSource, ZipPath
from bioimageio.spec.model import AnyModelDescr, v0_4, v0_5
from bioimageio.spec.model.v0_4 import CallableFromDepencency, CallableFromFile
from bioimageio.spec.model.v0_5 import (
    ArchitectureFromFileDescr,
    ArchitectureFromLibraryDescr,
    ParameterizedSize_N,
)
from bioimageio.spec.utils import load_array

from .axis import Axis, AxisId, AxisInfo, AxisLike, PerAxis
from .block_meta import split_multiple_shapes_into_blocks
from .common import Halo, MemberId, PerMember, SampleId, TotalNumberOfBlocks
from .io import load_tensor
from .sample import (
    LinearSampleAxisTransform,
    Sample,
    SampleBlockMeta,
    sample_block_meta_generator,
)
from .stat_measures import Stat
from .tensor import Tensor

TensorSource = Union[Tensor, xr.DataArray, NDArray[Any], Path]


def import_callable(
    node: Union[
        ArchitectureFromFileDescr,
        ArchitectureFromLibraryDescr,
        CallableFromDepencency,
        CallableFromFile,
    ],
    /,
    **kwargs: Unpack[HashKwargs],
) -> Callable[..., Any]:
    """import a callable (e.g. a torch.nn.Module) from a spec node describing it"""
    if isinstance(node, CallableFromDepencency):
        module = importlib.import_module(node.module_name)
        c = getattr(module, str(node.callable_name))
    elif isinstance(node, ArchitectureFromLibraryDescr):
        module = importlib.import_module(node.import_from)
        c = getattr(module, str(node.callable))
    elif isinstance(node, CallableFromFile):
        c = _import_from_file_impl(node.source_file, str(node.callable_name), **kwargs)
    elif isinstance(node, ArchitectureFromFileDescr):
        c = _import_from_file_impl(node.source, str(node.callable), sha256=node.sha256)
    else:
        assert_never(node)

    if not callable(c):
        raise ValueError(f"{node} (imported: {c}) is not callable")

    return c


tmp_dirs_in_use: List[TemporaryDirectory[str]] = []
"""keep global reference to temporary directories created during import to delay cleanup"""


def _import_from_file_impl(
    source: FileSource, callable_name: str, **kwargs: Unpack[HashKwargs]
):
    src_descr = FileDescr(source=source, **kwargs)
    # ensure sha is valid even if perform_io_checks=False
    # or the source has changed since last sha computation
    src_descr.validate_sha256(force_recompute=True)
    assert src_descr.sha256 is not None
    source_sha = src_descr.sha256

    reader = src_descr.get_reader()
    # make sure we have unique module name
    module_name = f"{reader.original_file_name.split('.')[0]}_{source_sha}"

    # make sure we have a unique and valid module name
    if not module_name.isidentifier():
        module_name = f"custom_module_{source_sha}"
        assert module_name.isidentifier(), module_name

    source_bytes = reader.read()

    module = sys.modules.get(module_name)
    if module is None:
        try:
            td_kwargs: Dict[str, Any] = (
                dict(ignore_cleanup_errors=True) if sys.version_info >= (3, 10) else {}
            )
            if sys.version_info >= (3, 12):
                td_kwargs["delete"] = False

            tmp_dir = TemporaryDirectory(**td_kwargs)
            # keep global ref to tmp_dir to delay cleanup until program exit
            # TODO: remove for py >= 3.12, when delete=False works
            tmp_dirs_in_use.append(tmp_dir)

            module_path = Path(tmp_dir.name) / module_name
            if reader.original_file_name.endswith(".zip") or is_zipfile(reader):
                module_path.mkdir()
                ZipFile(reader).extractall(path=module_path)
            else:
                module_path = module_path.with_suffix(".py")
                _ = module_path.write_bytes(source_bytes)

            importlib_spec = importlib.util.spec_from_file_location(
                module_name, str(module_path)
            )

            if importlib_spec is None:
                raise ImportError(f"Failed to import {source}")

            module = importlib.util.module_from_spec(importlib_spec)

            sys.modules[module_name] = module  # cache this module

            assert importlib_spec.loader is not None
            importlib_spec.loader.exec_module(module)

        except Exception as e:
            del sys.modules[module_name]
            raise ImportError(f"Failed to import {source}") from e

    try:
        callable_attr = getattr(module, callable_name)
    except AttributeError as e:
        raise AttributeError(
            f"Imported custom module from {source} has no `{callable_name}` attribute."
        ) from e
    except Exception as e:
        raise AttributeError(
            f"Failed to access `{callable_name}` attribute from custom module imported from {source} ."
        ) from e

    else:
        return callable_attr


def get_axes_infos(
    io_descr: Union[
        v0_4.InputTensorDescr,
        v0_4.OutputTensorDescr,
        v0_5.InputTensorDescr,
        v0_5.OutputTensorDescr,
    ],
) -> List[AxisInfo]:
    """get a unified, simplified axis representation from spec axes"""
    ret: List[AxisInfo] = []
    for a in io_descr.axes:
        if isinstance(a, v0_5.AxisBase):
            ret.append(AxisInfo.create(Axis(id=a.id, type=a.type)))
        else:
            assert a in ("b", "i", "t", "c", "z", "y", "x")
            ret.append(AxisInfo.create(a))

    return ret


def get_member_id(
    tensor_description: Union[
        v0_4.InputTensorDescr,
        v0_4.OutputTensorDescr,
        v0_5.InputTensorDescr,
        v0_5.OutputTensorDescr,
    ],
) -> MemberId:
    """get the normalized tensor ID, usable as a sample member ID"""

    if isinstance(tensor_description, (v0_4.InputTensorDescr, v0_4.OutputTensorDescr)):
        return MemberId(tensor_description.name)
    elif isinstance(
        tensor_description, (v0_5.InputTensorDescr, v0_5.OutputTensorDescr)
    ):
        return tensor_description.id
    else:
        assert_never(tensor_description)


def get_member_ids(
    tensor_descriptions: Sequence[
        Union[
            v0_4.InputTensorDescr,
            v0_4.OutputTensorDescr,
            v0_5.InputTensorDescr,
            v0_5.OutputTensorDescr,
        ]
    ],
) -> List[MemberId]:
    """get normalized tensor IDs to be used as sample member IDs"""
    return [get_member_id(descr) for descr in tensor_descriptions]


def get_test_input_sample(model: AnyModelDescr) -> Sample:
    return _get_test_sample(
        model.inputs,
        model.test_inputs if isinstance(model, v0_4.ModelDescr) else model.inputs,
    )


get_test_inputs = get_test_input_sample
"""DEPRECATED: use `get_test_input_sample` instead"""


def get_test_output_sample(model: AnyModelDescr) -> Sample:
    """returns a model's test output sample"""
    return _get_test_sample(
        model.outputs,
        model.test_outputs if isinstance(model, v0_4.ModelDescr) else model.outputs,
    )


get_test_outputs = get_test_output_sample
"""DEPRECATED: use `get_test_input_sample` instead"""


def _get_test_sample(
    tensor_descrs: Sequence[
        Union[
            v0_4.InputTensorDescr,
            v0_4.OutputTensorDescr,
            v0_5.InputTensorDescr,
            v0_5.OutputTensorDescr,
        ]
    ],
    test_sources: Sequence[Union[FileSource, v0_5.TensorDescr]],
) -> Sample:
    """returns a model's input/output test sample"""
    member_ids = get_member_ids(tensor_descrs)
    arrays: List[NDArray[Any]] = []
    for src in test_sources:
        if isinstance(src, (v0_5.InputTensorDescr, v0_5.OutputTensorDescr)):
            if src.test_tensor is None:
                raise ValueError(
                    f"Model input '{src.id}' has no test tensor defined, cannot create test sample."
                )
            arrays.append(load_array(src.test_tensor))
        else:
            arrays.append(load_array(src))

    axes = [get_axes_infos(t) for t in tensor_descrs]
    return Sample(
        members={
            m: Tensor.from_numpy(arr, dims=ax)
            for m, arr, ax in zip(member_ids, arrays, axes)
        },
        stat={},
        id="test-sample",
    )


class IO_SampleBlockMeta(NamedTuple):
    input: SampleBlockMeta
    output: SampleBlockMeta


def get_input_halo(model: v0_5.ModelDescr, output_halo: PerMember[PerAxis[Halo]]):
    """returns which halo input tensors need to be divided into blocks with, such that
    `output_halo` can be cropped from their outputs without introducing gaps."""
    input_halo: Dict[MemberId, Dict[AxisId, Halo]] = {}
    outputs = {t.id: t for t in model.outputs}
    all_tensors = {**{t.id: t for t in model.inputs}, **outputs}

    for t, th in output_halo.items():
        axes = {a.id: a for a in outputs[t].axes}

        for a, ah in th.items():
            s = axes[a].size
            if not isinstance(s, v0_5.SizeReference):
                raise ValueError(
                    f"Unable to map output halo for {t}.{a} to an input axis"
                )

            axis = axes[a]
            ref_axis = {a.id: a for a in all_tensors[s.tensor_id].axes}[s.axis_id]

            total_output_halo = sum(ah)
            total_input_halo = total_output_halo * axis.scale / ref_axis.scale
            assert (
                total_input_halo == int(total_input_halo) and total_input_halo % 2 == 0
            )
            input_halo.setdefault(s.tensor_id, {})[a] = Halo(
                int(total_input_halo // 2), int(total_input_halo // 2)
            )

    return input_halo


def get_block_transform(
    model: v0_5.ModelDescr,
) -> PerMember[PerAxis[Union[LinearSampleAxisTransform, int]]]:
    """returns how a model's output tensor shapes relates to its input shapes"""
    ret: Dict[MemberId, Dict[AxisId, Union[LinearSampleAxisTransform, int]]] = {}
    batch_axis_trf = None
    for ipt in model.inputs:
        for a in ipt.axes:
            if a.type == "batch":
                batch_axis_trf = LinearSampleAxisTransform(
                    axis=a.id, scale=1, offset=0, member=ipt.id
                )
                break
        if batch_axis_trf is not None:
            break
    axis_scales = {
        t.id: {a.id: a.scale for a in t.axes}
        for t in chain(model.inputs, model.outputs)
    }
    for out in model.outputs:
        new_axes: Dict[AxisId, Union[LinearSampleAxisTransform, int]] = {}
        for a in out.axes:
            if a.size is None:
                assert a.type == "batch"
                if batch_axis_trf is None:
                    raise ValueError(
                        "no batch axis found in any input tensor, but output tensor"
                        + f" '{out.id}' has one."
                    )
                s = batch_axis_trf
            elif isinstance(a.size, int):
                s = a.size
            elif isinstance(a.size, v0_5.DataDependentSize):
                s = -1
            elif isinstance(a.size, v0_5.SizeReference):
                s = LinearSampleAxisTransform(
                    axis=a.size.axis_id,
                    scale=axis_scales[a.size.tensor_id][a.size.axis_id] / a.scale,
                    offset=a.size.offset,
                    member=a.size.tensor_id,
                )
            else:
                assert_never(a.size)

            new_axes[a.id] = s

        ret[out.id] = new_axes

    return ret


def get_io_sample_block_metas(
    model: v0_5.ModelDescr,
    input_sample_shape: PerMember[PerAxis[int]],
    ns: Mapping[Tuple[MemberId, AxisId], ParameterizedSize_N],
    batch_size: int = 1,
) -> Tuple[TotalNumberOfBlocks, Iterable[IO_SampleBlockMeta]]:
    """returns an iterable yielding meta data for corresponding input and output samples"""
    if not isinstance(model, v0_5.ModelDescr):
        raise TypeError(f"get_block_meta() not implemented for {type(model)}")

    block_axis_sizes = model.get_axis_sizes(ns=ns, batch_size=batch_size)
    input_block_shape = {
        t: {aa: s for (tt, aa), s in block_axis_sizes.inputs.items() if tt == t}
        for t in {tt for tt, _ in block_axis_sizes.inputs}
    }
    output_halo = {
        t.id: {
            a.id: Halo(a.halo, a.halo) for a in t.axes if isinstance(a, v0_5.WithHalo)
        }
        for t in model.outputs
    }
    input_halo = get_input_halo(model, output_halo)

    n_input_blocks, input_blocks = split_multiple_shapes_into_blocks(
        input_sample_shape, input_block_shape, halo=input_halo
    )
    block_transform = get_block_transform(model)
    return n_input_blocks, (
        IO_SampleBlockMeta(ipt, ipt.get_transformed(block_transform))
        for ipt in sample_block_meta_generator(
            input_blocks, sample_shape=input_sample_shape, sample_id=None
        )
    )


def get_tensor(
    src: Union[ZipPath, TensorSource],
    ipt: Union[v0_4.InputTensorDescr, v0_5.InputTensorDescr],
):
    """helper to cast/load various tensor sources"""

    if isinstance(src, Tensor):
        return src
    elif isinstance(src, xr.DataArray):
        return Tensor.from_xarray(src)
    elif isinstance(src, np.ndarray):
        return Tensor.from_numpy(src, dims=get_axes_infos(ipt))
    else:
        return load_tensor(src, axes=get_axes_infos(ipt))


def create_sample_for_model(
    model: AnyModelDescr,
    *,
    stat: Optional[Stat] = None,
    sample_id: SampleId = None,
    inputs: Union[PerMember[TensorSource], TensorSource],
) -> Sample:
    """Create a sample from a single set of input(s) for a specific bioimage.io model

    Args:
        model: a bioimage.io model description
        stat: dictionary with sample and dataset statistics (may be updated in-place!)
        inputs: the input(s) constituting a single sample.
    """

    model_inputs = {get_member_id(d): d for d in model.inputs}
    if isinstance(inputs, collections.abc.Mapping):
        inputs = {MemberId(k): v for k, v in inputs.items()}
    elif len(model_inputs) == 1:
        inputs = {list(model_inputs)[0]: inputs}
    else:
        raise TypeError(
            f"Expected `inputs` to be a mapping with keys {tuple(model_inputs)}"
        )

    if unknown := {k for k in inputs if k not in model_inputs}:
        raise ValueError(f"Got unexpected inputs: {unknown}")

    if missing := {
        k
        for k, v in model_inputs.items()
        if k not in inputs and not (isinstance(v, v0_5.InputTensorDescr) and v.optional)
    }:
        raise ValueError(f"Missing non-optional model inputs: {missing}")

    return Sample(
        members={
            m: get_tensor(inputs[m], ipt)
            for m, ipt in model_inputs.items()
            if m in inputs
        },
        stat={} if stat is None else stat,
        id=sample_id,
    )


def load_sample_for_model(
    *,
    model: AnyModelDescr,
    paths: PerMember[Path],
    axes: Optional[PerMember[Sequence[AxisLike]]] = None,
    stat: Optional[Stat] = None,
    sample_id: Optional[SampleId] = None,
):
    """load a single sample from `paths` that can be processed by `model`"""

    if axes is None:
        axes = {}

    # make sure members are keyed by MemberId, not string
    paths = {MemberId(k): v for k, v in paths.items()}
    axes = {MemberId(k): v for k, v in axes.items()}

    model_inputs = {get_member_id(d): d for d in model.inputs}

    if unknown := {k for k in paths if k not in model_inputs}:
        raise ValueError(f"Got unexpected paths for {unknown}")

    if unknown := {k for k in axes if k not in model_inputs}:
        raise ValueError(f"Got unexpected axes hints for: {unknown}")

    members: Dict[MemberId, Tensor] = {}
    for m, p in paths.items():
        if m not in axes:
            axes[m] = get_axes_infos(model_inputs[m])
            logger.debug(
                "loading '{}' from {} with default input axes {} ",
                m,
                p,
                axes[m],
            )
        members[m] = load_tensor(p, axes[m])

    return Sample(
        members=members,
        stat={} if stat is None else stat,
        id=sample_id or tuple(sorted(paths.values())),
    )



import sys

from loguru import logger

logger.enable("bioimageio")

logger.remove()
_ = logger.add(
    sys.stderr,
    level="INFO",
    format="<green>{elapsed:}</green> | "
    + "<level>{level: <8}</level> | "
    + "<cyan>{module}</cyan> - <level>{message}</level>",
)

from .cli import Bioimageio


def main():
    cli = Bioimageio()  # pyright: ignore[reportCallIssue]
    cli.run()


if __name__ == "__main__":
    main()



# this file was modified from the generated
# https://github.com/pydata/xarray/blob/cf3655968b8b12cc0ecd28fb324e63fb94d5e7e2/xarray/core/_typed_ops.py
# TODO: should we generate this ourselves?
# TODO: test these magic methods
import operator
from typing import Any, Callable

from typing_extensions import Self
from xarray.core import nputils, ops


class MagicTensorOpsMixin:
    __slots__ = ()
    _Compatible = Any

    def _binary_op(
        self,
        other: _Compatible,
        f: Callable[[Any, Any], Any],
        reflexive: bool = False,
    ) -> Self:
        raise NotImplementedError

    def __add__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.add)

    def __sub__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.sub)

    def __mul__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.mul)

    def __pow__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.pow)

    def __truediv__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.truediv)

    def __floordiv__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.floordiv)

    def __mod__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.mod)

    def __and__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.and_)

    def __xor__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.xor)

    def __or__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.or_)

    def __lshift__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.lshift)

    def __rshift__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.rshift)

    def __lt__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.lt)

    def __le__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.le)

    def __gt__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.gt)

    def __ge__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.ge)

    def __eq__(self, other: _Compatible) -> Self:  # type: ignore[override]
        return self._binary_op(
            other, nputils.array_eq  # pyright: ignore[reportUnknownArgumentType]
        )

    def __ne__(self, other: _Compatible) -> Self:  # type: ignore[override]
        return self._binary_op(
            other, nputils.array_ne  # pyright: ignore[reportUnknownArgumentType]
        )

    # When __eq__ is defined but __hash__ is not, then an object is unhashable,
    # and it should be declared as follows:
    __hash__: None  # type:ignore[assignment]

    def __radd__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.add, reflexive=True)

    def __rsub__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.sub, reflexive=True)

    def __rmul__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.mul, reflexive=True)

    def __rpow__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.pow, reflexive=True)

    def __rtruediv__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.truediv, reflexive=True)

    def __rfloordiv__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.floordiv, reflexive=True)

    def __rmod__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.mod, reflexive=True)

    def __rand__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.and_, reflexive=True)

    def __rxor__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.xor, reflexive=True)

    def __ror__(self, other: _Compatible) -> Self:
        return self._binary_op(other, operator.or_, reflexive=True)

    def _inplace_binary_op(
        self, other: _Compatible, f: Callable[[Any, Any], Any]
    ) -> Self:
        raise NotImplementedError

    def __iadd__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.iadd)

    def __isub__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.isub)

    def __imul__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.imul)

    def __ipow__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.ipow)

    def __itruediv__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.itruediv)

    def __ifloordiv__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.ifloordiv)

    def __imod__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.imod)

    def __iand__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.iand)

    def __ixor__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.ixor)

    def __ior__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.ior)

    def __ilshift__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.ilshift)

    def __irshift__(self, other: _Compatible) -> Self:
        return self._inplace_binary_op(other, operator.irshift)

    def _unary_op(self, f: Callable[[Any], Any], *args: Any, **kwargs: Any) -> Self:
        raise NotImplementedError

    def __neg__(self) -> Self:
        return self._unary_op(operator.neg)

    def __pos__(self) -> Self:
        return self._unary_op(operator.pos)

    def __abs__(self) -> Self:
        return self._unary_op(operator.abs)

    def __invert__(self) -> Self:
        return self._unary_op(operator.invert)

    def round(self, *args: Any, **kwargs: Any) -> Self:
        return self._unary_op(
            ops.round_, *args, **kwargs  # pyright: ignore[reportUnknownArgumentType]
        )

    def argsort(self, *args: Any, **kwargs: Any) -> Self:
        return self._unary_op(
            ops.argsort, *args, **kwargs  # pyright: ignore[reportUnknownArgumentType]
        )

    def conj(self, *args: Any, **kwargs: Any) -> Self:
        return self._unary_op(
            ops.conj, *args, **kwargs  # pyright: ignore[reportUnknownArgumentType]
        )

    def conjugate(self, *args: Any, **kwargs: Any) -> Self:
        return self._unary_op(
            ops.conjugate, *args, **kwargs  # pyright: ignore[reportUnknownArgumentType]
        )

    __add__.__doc__ = operator.add.__doc__
    __sub__.__doc__ = operator.sub.__doc__
    __mul__.__doc__ = operator.mul.__doc__
    __pow__.__doc__ = operator.pow.__doc__
    __truediv__.__doc__ = operator.truediv.__doc__
    __floordiv__.__doc__ = operator.floordiv.__doc__
    __mod__.__doc__ = operator.mod.__doc__
    __and__.__doc__ = operator.and_.__doc__
    __xor__.__doc__ = operator.xor.__doc__
    __or__.__doc__ = operator.or_.__doc__
    __lshift__.__doc__ = operator.lshift.__doc__
    __rshift__.__doc__ = operator.rshift.__doc__
    __lt__.__doc__ = operator.lt.__doc__
    __le__.__doc__ = operator.le.__doc__
    __gt__.__doc__ = operator.gt.__doc__
    __ge__.__doc__ = operator.ge.__doc__
    __eq__.__doc__ = nputils.array_eq.__doc__
    __ne__.__doc__ = nputils.array_ne.__doc__
    __radd__.__doc__ = operator.add.__doc__
    __rsub__.__doc__ = operator.sub.__doc__
    __rmul__.__doc__ = operator.mul.__doc__
    __rpow__.__doc__ = operator.pow.__doc__
    __rtruediv__.__doc__ = operator.truediv.__doc__
    __rfloordiv__.__doc__ = operator.floordiv.__doc__
    __rmod__.__doc__ = operator.mod.__doc__
    __rand__.__doc__ = operator.and_.__doc__
    __rxor__.__doc__ = operator.xor.__doc__
    __ror__.__doc__ = operator.or_.__doc__
    __iadd__.__doc__ = operator.iadd.__doc__
    __isub__.__doc__ = operator.isub.__doc__
    __imul__.__doc__ = operator.imul.__doc__
    __ipow__.__doc__ = operator.ipow.__doc__
    __itruediv__.__doc__ = operator.itruediv.__doc__
    __ifloordiv__.__doc__ = operator.ifloordiv.__doc__
    __imod__.__doc__ = operator.imod.__doc__
    __iand__.__doc__ = operator.iand.__doc__
    __ixor__.__doc__ = operator.ixor.__doc__
    __ior__.__doc__ = operator.ior.__doc__
    __ilshift__.__doc__ = operator.ilshift.__doc__
    __irshift__.__doc__ = operator.irshift.__doc__
    __neg__.__doc__ = operator.neg.__doc__
    __pos__.__doc__ = operator.pos.__doc__
    __abs__.__doc__ = operator.abs.__doc__
    __invert__.__doc__ = operator.invert.__doc__



"""
.. include:: ../../README.md
"""
# ruff: noqa: E402

__version__ = "0.9.1"
from loguru import logger

logger.disable("bioimageio.core")

from bioimageio.spec import (
    ValidationSummary,
    build_description,
    dump_description,
    load_dataset_description,
    load_description,
    load_description_and_validate_format_only,
    load_model_description,
    save_bioimageio_package,
    save_bioimageio_package_as_folder,
    save_bioimageio_yaml_only,
    validate_format,
)

from . import (
    axis,
    block_meta,
    cli,
    commands,
    common,
    digest_spec,
    io,
    model_adapters,
    prediction,
    proc_ops,
    proc_setup,
    sample,
    stat_calculators,
    stat_measures,
    tensor,
)
from ._prediction_pipeline import PredictionPipeline, create_prediction_pipeline
from ._resource_tests import (
    enable_determinism,
    load_description_and_test,
    test_description,
    test_model,
)
from ._settings import settings
from .axis import Axis, AxisId
from .backends import create_model_adapter
from .block_meta import BlockMeta
from .common import MemberId
from .prediction import predict, predict_many
from .sample import Sample
from .stat_calculators import compute_dataset_measures
from .stat_measures import Stat
from .tensor import Tensor
from .weight_converters import add_weights

# aliases
test_resource = test_description
"""alias of `test_description`"""
load_resource = load_description
"""alias of `load_description`"""
load_model = load_model_description
"""alias of `load_model_description`"""

__all__ = [
    "__version__",
    "add_weights",
    "axis",
    "Axis",
    "AxisId",
    "block_meta",
    "BlockMeta",
    "build_description",
    "cli",
    "commands",
    "common",
    "compute_dataset_measures",
    "create_model_adapter",
    "create_prediction_pipeline",
    "digest_spec",
    "dump_description",
    "enable_determinism",
    "io",
    "load_dataset_description",
    "load_description_and_test",
    "load_description_and_validate_format_only",
    "load_description",
    "load_model_description",
    "load_model",
    "load_resource",
    "MemberId",
    "model_adapters",
    "predict_many",
    "predict",
    "prediction",
    "PredictionPipeline",
    "proc_ops",
    "proc_setup",
    "sample",
    "Sample",
    "save_bioimageio_package_as_folder",
    "save_bioimageio_package",
    "save_bioimageio_yaml_only",
    "settings",
    "stat_calculators",
    "stat_measures",
    "Stat",
    "tensor",
    "Tensor",
    "test_description",
    "test_model",
    "test_resource",
    "validate_format",
    "ValidationSummary",
]
